# -*- coding: utf-8 -*-
'''utils/progressbar.py'''

from __future__ import print_function, division

from time import time, localtime
from platform import system

from .func import get_stdout, term_size

def _pretty_time(my_time, ampm):
    # (my_time: float, ampm: bool) -> str:
    new_time = localtime(my_time)

    hours = new_time.tm_hour
    minutes = new_time.tm_min

    if(ampm):
        if(hours == 0):
            hours = 12
        if(hours > 12):
            hours -= 12
        ampm = 'PM' if new_time.tm_hour >= 12 else 'AM'
        return '{:02}:{:02} {}'.format(hours, minutes, ampm)
    return '{:02}:{:02}'.format(hours, minutes)

def bar(termsize, title, done, togo, percent, new_time):
    bar = '  ⏳{}: [{}{}] {}% done ETA {}'.format(title, done, togo, percent, new_time)
    if(len(bar) > termsize - 2):
        bar = bar[:termsize - 2]
    else:
        bar += ' ' * (termsize - len(bar) - 4)
    try:
        print(bar, end='\r', flush=True)
    except TypeError:
        print(bar, end='\r')


class ProgressBar():
    def __init__(self, total, title='Please wait', machineReadable=False, hide=False):

        self.total = total
        self.beginTime = time()
        self.title = title
        self.len_title = len(title)
        self.machine = machineReadable
        self.hide = hide
        self.ampm = True

        if(system() == 'Darwin' and not self.machine):
            try:
                dateFormat = get_stdout(['defaults', 'read',
                    'com.apple.menuextra.clock', 'DateFormat'])
                self.ampm = 'a' in dateFormat
            except FileNotFoundError:
                pass

        self.allow_unicode = True
        try:
            self.tick(0)
        except UnicodeEncodeError:
            newTime = _pretty_time(self.beginTime, self.ampm)
            print('   0% done ETA {}'.format(newTime))
            self.allow_unicode = False

    def tick(self, index):

        if(self.hide):
            return

        percentDone = min(100, round((index+1) / self.total * 100, 1))
        if(percentDone == 0): # Prevent dividing by zero.
            percentPerSec = 0
        else:
            percentPerSec = (time() - self.beginTime) / percentDone

        new_time = _pretty_time(self.beginTime + (percentPerSec * 100), self.ampm)

        if(self.machine):
            index = min(index, self.total)
            raw = int(self.beginTime + (percentPerSec * 100))
            print('{}~{}~{}~{}~{}'.format(
                self.title, index, self.total, self.beginTime, raw),
                end='\r', flush=True)
            return

        termsize = term_size().columns

        if(self.allow_unicode):
            bar_len = max(1, termsize - (self.len_title + 50))
            done = int(round(percentDone / (100 / bar_len)))
            togo = '░' * int(bar_len - done)
            bar(termsize, self.title, '█' * done, togo, percentDone, new_time)
        else:
            print('   {}% done ETA {}'.format(percentDone, new_time))



'''subcommands/create.py'''

def create_options(parser):
    parser.add_argument('--frame_rate', '-fps', '-r', type=float, default=30.0,
        help='set the framerate for the output video.')
    parser.add_argument('--duration', '-d', type=int, default=10,
        help='set the length of the video (in seconds).')
    parser.add_argument('--width', type=int, default=1280,
        help='set the pixel width of the video.')
    parser.add_argument('--height', type=int, default=720,
        help='set the pixel height of the video.')
    parser.add_argument('--output_file', '--output', '-o', type=str,
        default='testsrc.mp4')
    parser.add_argument('--ffmpeg_location', default=None,
        help='point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    parser.add_argument('(input)', nargs='*',
        help='the template')
    return parser

def create(sys_args=None):
    import os
    import sys
    import time

    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.log import Log
    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('create', auto_editor.version,
        description='Generate simple media.')
    parser = create_options(parser)

    log = Log()
    args = parser.parse_args(sys.argv[2:], Log(), 'create')

    dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    ffmpeg = FFmpeg(dir_path, args.ffmpeg_location, args.my_ffmpeg, False, log)

    theme = args.input
    output = args.output_file
    fps = args.frame_rate

    if(len(theme) == 0):
        log.error('You must put a theme!')
    if(len(theme) > 1):
        log.error('Only one theme at a time.')
    theme = theme[0]

    try:
        os.remove(output)
    except FileNotFoundError:
        pass

    if(theme == 'test'):
        # Create sine wav.
        ffmpeg.run(['-f', 'lavfi', '-i', 'sine=frequency=1000:duration=0.2', 'short.wav'])
        ffmpeg.run(['-i', 'short.wav', '-af', 'apad', '-t', '1', 'beep.wav']) # Pad audio.

        # Generate video with no audio.
        ffmpeg.run(['-f', 'lavfi', '-i', 'testsrc=duration={}:size={}x{}:rate={}'.format(
            args.duration, args.width, args.height, fps), '-pix_fmt', 'yuv420p', output])

        # Add empty audio channel to video.
        ffmpeg.run(['-f', 'lavfi', '-i', 'anullsrc=channel_layout=stereo:sample_rate=44100',
            '-i', output, '-c:v', 'copy', '-c:a', 'aac', '-shortest', 'pre' + output])

        # Mux Video with repeating audio.
        ffmpeg.run(['-i', 'pre' + output, '-filter_complex',
            'amovie=beep.wav:loop=0,asetpts=N/SR/TB[aud];[0:a][aud]amix[a]',
            '-map', '0:v', '-map', '[a]', '-c:v', 'copy', '-c:a', 'aac', '-b:a', '256k',
            '-shortest', output])

        time.sleep(1)
        os.remove('short.wav')
        os.remove('beep.wav')
        os.remove('pre' + output)

    if(theme in ['white', 'black']):
        ffmpeg.run(['-f', 'lavfi', '-i', 'color=size={}x{}:rate={}:color={}'.format(
            args.width, args.height, fps, theme), '-t', str(args.duration), output])

if(__name__ == '__main__'):
    create()



'''formats/utils.py'''

def safe_mkdir(path):
    from shutil import rmtree
    from os import mkdir
    try:
        mkdir(path)
    except OSError:
        rmtree(path)
        mkdir(path)
    return path

def get_width_height(inp):
    if(len(inp.video_streams) == 0):
        return None, None
    else:
        return inp.video_streams[0]['width'], inp.video_streams[0]['height']

def indent(base: int, *args: str) -> str:
    r = ''
    for line in args:
        r += ('\t' * base) + line + '\n'
    return r



'''interpolate.py'''

import math

def linear(x: int, y: int, n: int) -> list:
    b = [x]
    step = (y - x) / n
    incre = x
    for _ in range(n - 2):
        incre += step
        b.append(incre)

    b.append(y)
    return b

# See how these formulas are derived:
# - https://www.desmos.com/calculator/jj4tociyb4

def sine(x, y, n: int) -> list:
    # slow -> fast -> slow

    b = [x]
    incre = 0
    for _ in range(n - 2):
        incre += math.pi / n

        val = ((y - x)/2) * math.sin(incre - (math.pi / 2)) + ((y - x)/2) + x
        b.append(val)

    b.append(y)
    return b

def start_sine(x, y, n: int) -> list:
    # slow -> fast

    b = [x]
    incre = 0
    for _ in range(n - 2):
        incre += math.pi / n
        val = ((y - x)/2) * math.sin(incre - (math.pi / 2)) + ((y - x)/2) + x
        b.append(val)

    b.append(y)
    return b

def end_sine(x, y, n: int) -> list:
    # fast -> slow

    b = [x]
    incre = 0
    for _ in range(n - 2):
        incre += (math.pi / 2) / n

        val = x + math.sin(incre) * (y - x)
        b.append(val)

    b.append(y)
    return b

def interpolate(x, y, n, log, method='linear') -> list:
    if(method == 'linear'):
        return linear(x, y, n)
    elif(method == 'sine'):
        return sine(x, y, n)
    elif(method == 'start_sine'):
        return start_sine(x, y, n)
    elif(method == 'end_sine'):
        return end_sine(x, y, n)
    else:
        log.error("Method: {} isn't implemented.".format(method))



'''ffwrapper.py'''

from __future__ import print_function

# Internal Libraries
import subprocess
import os.path
from platform import system

# Included Libraries
from auto_editor.utils.func import get_stdout

def _set_ff_path(dirpath, ff_location, my_ffmpeg):
    from platform import system
    if(ff_location is not None):
        return ff_location
    if(my_ffmpeg or system() not in ['Windows', 'Darwin']):
        return 'ffmpeg'
    program = 'ffmpeg' if system() == 'Darwin' else 'ffmpeg.exe'
    return os.path.join(dirpath, 'ffmpeg', system(), program)

def _test_path(path, log):
    try:
        get_stdout([path, '-h'])
    except FileNotFoundError:
        if(system() == 'Darwin'):
            log.error('No ffmpeg found, download via homebrew or restore the '\
                'included binary.')
        if(system() == 'Windows'):
            log.error('No ffmpeg found, download ffmpeg with your favorite package '\
                'manager (ex chocolatey), or restore the included binary.')

        log.error('ffmpeg must be on PATH. Download ffmpeg by running:\n'\
        '  sudo apt-get install libavformat-dev libavfilter-dev libavdevice-dev ffmpeg'\
        '\nOr something similar depending on your distro.')
    return path

class FFmpeg():
    def __init__(self, dirpath, ff_location, my_ffmpeg, debug, log):
        _path = _set_ff_path(dirpath, ff_location, my_ffmpeg)
        self.log = log
        self.path = _test_path(_path, log)
        self.FFdebug = debug

    def getPath(self):
        return self.path

    def run(self, cmd):
        cmd = [self.path, '-y'] + cmd
        if(self.FFdebug):
            cmd.extend(['-hide_banner'])
            print(cmd)
        else:
            cmd.extend(['-nostats', '-loglevel', 'error'])
        subprocess.call(cmd)

    def file_info(self, path):
        import re

        class File:
            def __repr__(self):
                return str(self.__dict__)

        file = File()

        setattr(file, 'path', path)
        setattr(file, 'abspath', os.path.abspath(path))
        setattr(file, 'basename', os.path.basename(path))
        setattr(file, 'dirname', os.path.dirname(os.path.abspath(path)))

        f_name, f_ext = os.path.splitext(path)

        setattr(file, 'name', f_name)
        setattr(file, 'ext', f_ext)

        def regex_match(regex, text):
            match = re.search(regex, text)
            if(match):
                return match.groupdict()['match']
            return None

        info = get_stdout([self.path, '-hide_banner', '-i', path])

        setattr(file, 'duration', regex_match(r'Duration:\s(?P<match>[\d:.]+),', info))
        setattr(file, 'bitrate', regex_match(r'bitrate:\s(?P<match>\d+\skb\/s)', info))

        video_streams = []
        audio_streams = []
        subtitle_streams = []
        fps = None

        for line in info.split('\n'):
            if(re.search(r'Stream #', line)):
                s_data = {}
                if(re.search(r'Video:', line)):
                    s_data['width'] = regex_match(r'(?P<match>\d+)x\d+[\s,]', line)
                    s_data['height'] = regex_match(r'\d+x(?P<match>\d+)[\s,]', line)
                    s_data['codec'] = regex_match(r'Video:\s(?P<match>\w+)', line)
                    s_data['bitrate'] = regex_match(r'\s(?P<match>\d+\skb\/s)', line)
                    if(fps is None):
                        fps = regex_match(r'\s(?P<match>[\d\.]+)\stbr', line)
                    video_streams.append(s_data)

                elif(re.search(r'Audio:', line)):
                    s_data['codec'] = regex_match(r'Audio:\s(?P<match>\w+)', line)
                    s_data['samplerate'] = regex_match(r'(?P<match>\d+)\sHz', line)
                    s_data['bitrate'] = regex_match(r'\s(?P<match>\d+\skb\/s)', line)
                    audio_streams.append(s_data)

                elif(re.search(r'Subtitle:', line)):
                    s_data['lang'] = regex_match(r'Stream #\d:\d\((?P<match>\w+)\)', line)
                    subtitle_streams.append(s_data)

        setattr(file, 'fps', fps)
        setattr(file, 'video_streams', video_streams)
        setattr(file, 'audio_streams', audio_streams)
        setattr(file, 'subtitle_streams', subtitle_streams)
        return file

    def Popen(self, cmd):
        cmd = [self.path] + cmd
        if(self.FFdebug):
            return subprocess.Popen(cmd, stdout=subprocess.PIPE)
        return subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.DEVNULL)

    def pipe(self, cmd):
        cmd = [self.path, '-y'] + cmd

        if(self.FFdebug):
            print(cmd)
        output = get_stdout(cmd)
        if(self.FFdebug):
            print(output)

        return output

    def getVersion(self):
        _version = self.pipe(['-version']).split('\n')[0]
        _version = _version.replace('ffmpeg version', '').strip()
        return _version.split(' ')[0]



'''formats/json_cutlist.py'''

"""
Make a pre-edited file reference that can be inputted back into auto-editor.
"""

import os
import json

def read_json_cutlist(json_file, version, log):
    with open(json_file, 'r') as f:
        data = json.load(f)

    if(data['presets']['version'] != version):
        log.warning('This json file was generated using a different '\
            'version of auto-editor.')

    media_file = data['timeline']['media_file']

    if(not os.path.isfile(media_file)):
        log.error('Could not locate media file: {}'.format(media_file))

    speeds = data['presets']['speeds']
    chunks = data['timeline']['chunks']

    return media_file, chunks, speeds


def make_json_cutlist(media_file, out, version, chunks, speeds, log):
    if(not out.endswith('.json')):
        log.error('Output extension must be .json')

    data = {}
    data['presets'] = {
        'version': version,
        'speeds': speeds,
    }
    data['timeline']= {
        'media_file': os.path.abspath(media_file),
        'chunks': chunks,
    }

    with open(out, 'w') as outfile:
        json.dump(data, outfile, indent=4)



'''utils/types.py'''

from __future__ import print_function, division

import os.path
import sys

from .func import clean_list

def error(message):
    print('Error! {}'.format(message), file=sys.stderr)
    sys.exit(1)

def file_type(path):
    if(not os.path.isfile(path)):
        error('Auto-Editor could not find the file: {}'.format(path))
    return path

def float_type(num):
    if(num.endswith('%')):
        return float(num[:-1]) / 100
    return float(num)

def sample_rate_type(num):
    if(num.endswith(' Hz')):
        return int(num[:-3])
    if(num.endswith(' kHz')):
        return int(float(num[:-4]) * 1000)
    if(num.endswith('kHz')):
        return int(float(num[:-3]) * 1000)
    if(num.endswith('Hz')):
        return int(num[:-2])
    return int(num)

def frame_type(num):
    # (num: str) -> int | str:
    if(num.endswith('f')):
        return int(num[:-1])
    if(num.endswith('sec')):
        return num[:-3]
    if(num.endswith('secs')):
        return num[:-4]
    if(num.endswith('s')):
        return num[:-1]
    return int(num)

def comma_type(inp, min_args=1, max_args=None, name=''):
    inp = clean_list(inp.split(','), '\r\n\t')
    if(min_args > len(inp)):
        error('Too few comma arguments for {}.'.format(name))
    if(max_args is not None and len(inp) > max_args):
        error('Too many comma arguments for {}.'.format(name))
    return inp

def zoom_type(inp):
    return comma_type(inp, 3, 8, 'zoom_type')

def rect_type(inp):
    return comma_type(inp, 6, 8, 'rect_type')

def range_type(inp):
    return comma_type(inp, 2, 2, 'range_type')

def speed_range_type(inp):
    return comma_type(inp, 3, 3, 'speed_range_type')

def block_type(inp):
    return comma_type(inp, 1, None, 'block_type')



#!/usr/bin/env python3
'''__main__.py'''

# Python 2 Compatibility
from __future__ import print_function, absolute_import

# Internal Libraries
import os
import sys
import tempfile

# Included Libraries
import auto_editor
import auto_editor.vanparse as vanparse
import auto_editor.utils.func as usefulfunctions

from auto_editor.utils.func import fnone, append_filename
from auto_editor.utils.log import Log, Timer
from auto_editor.ffwrapper import FFmpeg

def set_output_name(path, making_data_file, args):
    dot_index = path.rfind('.')

    if(dot_index == -1):
        root = path
    else:
        root = path[:dot_index]

    if(args.export_as_json):
        return root + '.json'
    if(args.export_to_final_cut_pro):
        return root + '.fcpxml'
    if(args.export_to_shotcut):
        return root + '.mlt'
    if(making_data_file):
        return root + '.xml'
    if(args.export_as_audio):
        return root + '_ALTERED.wav'

    ext = path[dot_index:]
    return root + '_ALTERED' + ext


def main_options(parser):
    from auto_editor.utils.types import (file_type, float_type, sample_rate_type,
        frame_type, zoom_type, rect_type, range_type, speed_range_type, block_type)

    parser.add_argument('progressOps', nargs=0, action='grouping')
    parser.add_argument('--machine_readable_progress', action='store_true',
        group='progressOps',
        help='set progress bar that is easier to parse.')
    parser.add_argument('--no_progress', action='store_true',
        group='progressOps',
        help='do not display any progress at all.')

    parser.add_argument('metadataOps', nargs=0, action='grouping')
    parser.add_argument('--force_fps_to', type=float, group='metadataOps',
        help='manually set the fps value for the input video if detection fails.')

    parser.add_argument('motionOps', nargs=0, action='grouping')
    parser.add_argument('--dilates', '-d', type=int, default=2, range='0 to 5',
        group='motionOps',
        help='set how many times a frame is dilated before being compared.')
    parser.add_argument('--width', '-w', type=int, default=400, range='1 to Infinity',
        group='motionOps',
        help="scale the frame to this width before being compared.")
    parser.add_argument('--blur', '-b', type=int, default=21, range='0 to Infinity',
        group='motionOps',
        help='set the strength of the blur applied to a frame before being compared.')

    parser.add_argument('urlOps', nargs=0, action='grouping')
    parser.add_argument('--format', type=str, group='urlOps',
        default='bestvideo[ext=mp4]+bestaudio[ext=m4a]/mp4',
        help='the format youtube-dl uses to when downloading a url.')
    parser.add_argument('--output_dir', type=str, group='urlOps',
        default=None,
        help='the directory where the downloaded file is placed.')
    parser.add_argument('--limit_rate', '-rate', default='3m',
        help='maximum download rate in bytes per second (50k, 4.2m)')
    parser.add_argument('--id', type=str, default=None, group='urlOps',
        help='manually set the YouTube ID the video belongs to.')
    parser.add_argument('--block', type=block_type, group='urlOps',
        help='mark all sponsors sections as silent.',
        extra='Only for YouTube urls. This uses the SponsorBlock api.\n'\
            'Choices can include: sponsor intro outro selfpromo interaction music_offtopic')
    parser.add_argument('--download_archive', type=file_type, default=None, group='urlOps',
        help='Download only videos not listed in the archive file. Record the IDs of' \
             ' all downloaded videos in it')
    parser.add_argument('--cookies', type=file_type, default=None, group='urlOps',
        help='The file to read cookies from and dump the cookie jar in.')
    parser.add_argument('--check_certificate', action='store_true', group='urlOps',
        help='check the website certificate before downloading.')

    parser.add_argument('exportMediaOps', nargs=0, action='grouping')
    parser.add_argument('--video_bitrate', '-vb', default='unset', group='exportMediaOps',
        help='set the number of bits per second for video.')
    parser.add_argument('--audio_bitrate', '-ab', default='unset', group='exportMediaOps',
        help='set the number of bits per second for audio.')
    parser.add_argument('--sample_rate', '-r', type=sample_rate_type,
        group='exportMediaOps',
        help='set the sample rate of the input and output videos.')
    parser.add_argument('--video_codec', '-vcodec', default='uncompressed',
        group='exportMediaOps',
        help='set the video codec for the output media file.')
    parser.add_argument('--audio_codec', '-acodec', group='exportMediaOps',
        help='set the audio codec for the output media file.')
    parser.add_argument('--preset', '-p', default='unset', group='exportMediaOps',
        choices=['ultrafast', 'superfast', 'veryfast', 'faster', 'fast', 'medium',
            'slow', 'slower', 'veryslow', 'unset'],
        help='set the preset for ffmpeg to help save file size or increase quality.')
    parser.add_argument('--tune', '-t', default='unset', group='exportMediaOps',
        choices=['film', 'animation', 'grain', 'stillimage', 'fastdecode',
            'zerolatency', 'none', 'unset'],
        help='set the tune for ffmpeg to compress video better in certain circumstances.')
    parser.add_argument('--constant_rate_factor', '-crf', default='unset',
        group='exportMediaOps', range='0 to 51',
        help='set the quality for video using the crf method.')

    parser.add_argument('--render', default='auto', choices=['av', 'opencv', 'auto'],
        help='choice which method to render video.')
    parser.add_argument('--scale', type=float_type, default=1,
        help='scale the output media file by a certain factor.')
    parser.add_argument('--combine_files', action='store_true',
        help='combine all input files into one before editing.')

    parser.add_argument('--zoom', type=zoom_type, nargs='*',
        help='set when and how a zoom will occur.',
        extra='The arguments are: start,end,start_zoom,end_zoom,x,y,inter,hold' \
            '\nThere must be at least 3 comma args. x and y default to centerX and centerY' \
            '\nThe default interpolation is linear.')
    parser.add_argument('--rectangle', type=rect_type, nargs='*',
        help='overlay a rectangle shape on the video.',
        extra='The arguments are: start,end,x1,y1,x2,y2,color,thickness' \
            '\nThere must be at least 6 comma args. The rectangle is solid if' \
            ' thickness is not defined.\n The default color is #000.')

    parser.add_argument('--background', type=str, default='#000',
        help='set the color of the background that is visible when the video is moved.')

    parser.add_argument('--mark_as_loud', type=range_type, nargs='*',
        help='the range that will be marked as "loud".')
    parser.add_argument('--mark_as_silent', type=range_type, nargs='*',
        help='the range that will be marked as "silent".')
    parser.add_argument('--cut_out', type=range_type, nargs='*',
        help='the range of media that will be removed completely, regardless of the '\
            'value of silent speed.')
    parser.add_argument('--set_speed_for_range', type=speed_range_type, nargs='*',
        help='set an arbitrary speed for a given range.',
        extra='The arguments are: speed,start,end')

    parser.add_argument('--motion_threshold', type=float_type, default=0.02,
        range='0 to 1',
        help='how much motion is required to be considered "moving"')
    parser.add_argument('--edit_based_on', '--edit', default='audio',
        choices=['audio', 'motion', 'none', 'not_audio', 'not_motion', 'audio_or_motion',
            'audio_and_motion', 'audio_xor_motion', 'audio_and_not_motion',
            'not_audio_and_motion', 'not_audio_and_not_motion'],
        help='decide which method to use when making edits.')

    parser.add_argument('--cut_by_this_audio', '-ca', type=file_type,
        help="base cuts by this audio file instead of the video's audio.")
    parser.add_argument('--cut_by_this_track', '-ct', type=int, default=0,
        range='0 to the number of audio tracks minus one',
        help='base cuts by a different audio track in the video.')
    parser.add_argument('--cut_by_all_tracks', '-cat', action='store_true',
        help='combine all audio tracks into one before basing cuts.')
    parser.add_argument('--keep_tracks_seperate', action='store_true',
        help="don't combine audio tracks when exporting.")

    parser.add_argument('--export_as_audio', '-exa', action='store_true',
        help='export as a WAV audio file.')

    parser.add_argument('--export_to_premiere', '-exp', action='store_true',
        help='export as an XML file for Adobe Premiere Pro instead of making a media file.')
    parser.add_argument('--export_to_resolve', '-exr', action='store_true',
        help='export as an XML file for DaVinci Resolve instead of making a media file.')
    parser.add_argument('--export_to_final_cut_pro', '-exf', action='store_true',
        help='export as an XML file for Final Cut Pro instead of making a media file.')
    parser.add_argument('--export_to_shotcut', '-exs', action='store_true',
        help='export as an XML timeline file for Shotcut instead of making a media file.')
    parser.add_argument('--export_as_json', action='store_true',
        help='export as a JSON file that can be read by auto-editor later.')
    parser.add_argument('--export_as_clip_sequence', '-excs', action='store_true',
        help='export as multiple numbered media files.')

    parser.add_argument('--ffmpeg_location', default=None,
        help='set a custom path to the ffmpeg location.',
        extra='This takes precedence over --my_ffmpeg.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='use the ffmpeg on your PATH instead of the one packaged.',
        extra='this is equivalent to --ffmpeg_location ffmpeg.')
    parser.add_argument('--version', action='store_true',
        help='show which auto-editor you have.')
    parser.add_argument('--debug', '--verbose', '-d', action='store_true',
        help='show debugging messages and values.')
    parser.add_argument('--show_ffmpeg_debug', action='store_true',
        help='show ffmpeg progress and output.')
    parser.add_argument('--quiet', '-q', action='store_true',
        help='display less output.')

    parser.add_argument('--preview', action='store_true',
        help='show stats on how the input will be cut.')
    parser.add_argument('--no_open', action='store_true',
        help='do not open the file after editing is done.')
    parser.add_argument('--min_clip_length', '-mclip', type=frame_type, default=3,
        range='0 to Infinity',
        help='set the minimum length a clip can be. If a clip is too short, cut it.')
    parser.add_argument('--min_cut_length', '-mcut', type=frame_type, default=6,
        range='0 to Infinity',
        help="set the minimum length a cut can be. If a cut is too short, don't cut")

    parser.add_argument('--output_file', '--output', '-o', nargs='*',
        help='set the name(s) of the new output.')
    parser.add_argument('--silent_threshold', '-t', type=float_type, default=0.04,
        range='0 to 1',
        help='set the volume that frames audio needs to surpass to be "loud".')
    parser.add_argument('--silent_speed', '-s', type=float_type, default=99999,
        range='Any number. Values <= 0 or >= 99999 will be cut out.',
        help='set the speed that "silent" sections should be played at.')
    parser.add_argument('--video_speed', '--sounded_speed', '-v', type=float_type,
        default=1.00,
        range='Any number. Values <= 0 or >= 99999 will be cut out.',
        help='set the speed that "loud" sections should be played at.')
    parser.add_argument('--frame_margin', '--margin', '-m', type=frame_type, default=6,
        range='0 to Infinity',
        help='set how many "silent" frames of on either side of "loud" sections '\
            'be included.')

    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    parser.add_argument('(input)', nargs='*',
        help='the path to a file, folder, or url you want edited.')
    return parser


def get_chunks(inp, speeds, segment, fps, args, log, audioData=None, sampleRate=None):
    from auto_editor.cutting import (combine_audio_motion, combine_segment,
        apply_spacing_rules, apply_frame_margin, seconds_to_frames, cook, chunkify)

    frame_margin = seconds_to_frames(args.frame_margin, fps)
    min_clip = seconds_to_frames(args.min_clip_length, fps)
    min_cut = seconds_to_frames(args.min_cut_length, fps)

    def get_has_loud(inp, args, fps, audioData, sampleRate, log):
        from auto_editor.analyze import get_blank_list, audio_detection, motion_detection
        if(args.edit_based_on == 'none'):
            return get_blank_list(inp, audioData, sampleRate, fps)

        audioList, motionList = None, None

        if('audio' in args.edit_based_on):
            if(audioData is None):
                audioList = get_blank_list(inp, audioData, sampleRate, fps)
            else:
                audioList = audio_detection(audioData, sampleRate, args.silent_threshold,
                    fps, log)

        if('motion' in args.edit_based_on):
            if(len(inp.video_streams) == 0):
                motionList = get_blank_list(inp, audioData, sampleRate, fps)
            else:
                motionList = motion_detection(inp, args.motion_threshold, log,
                    width=args.width, dilates=args.dilates, blur=args.blur)

        if(audioList is not None and motionList is not None):
            if(len(audioList) > len(motionList)):
                audioList = audioList[:len(motionList)]
            elif(len(motionList) > len(audioList)):
                motionList = motionList[:len(audioList)]

        return combine_audio_motion(audioList, motionList, args.edit_based_on, log)

    has_loud = get_has_loud(inp, args, fps, audioData, sampleRate, log)
    has_loud = cook(has_loud, min_clip, min_cut)
    has_loud_length = len(has_loud)
    has_loud = apply_frame_margin(has_loud, has_loud_length, frame_margin)

    if(segment is not None):
        has_loud = combine_segment(has_loud, segment, fps)
    return apply_spacing_rules(has_loud, has_loud_length, min_clip, min_cut, speeds,
        fps, args, log)


def get_effects(audioData, sampleRate, fps, args, log):
    effects = []
    if(args.zoom != []):
        from auto_editor.cutting import applyZooms
        effects += applyZooms(args.zoom, audioData, sampleRate, fps, log)
    if(args.rectangle != []):
        from auto_editor.cutting import applyRects
        effects += applyRects(args.rectangle, audioData, sampleRate, fps, log)
    return effects

def edit_media(i, inp, ffmpeg, args, speeds, segment, exporting_to_editor, data_file,
    TEMP, log):
    chunks = None
    if(inp.ext == '.json'):
        from auto_editor.formats.make_json import read_json_cutlist

        INPUT_FILE, chunks, speeds = read_json_cutlist(inp.path, auto_editor.version, log)

        inp = ffmpeg.file_info(INPUT_FILE)

        newOutput = set_output_name(inp.path, data_file, args)
    else:
        newOutput = args.output_file[i]
        if(not os.path.isdir(inp.path) and '.' not in newOutput):
            newOutput = set_output_name(newOutput, data_file, args)

    log.debug('Input File: {}'.format(inp.path))
    log.debug('Output File: {}'.format(newOutput))

    if(os.path.isfile(newOutput) and inp.path != newOutput):
        log.debug('Removing already existing file: {}'.format(newOutput))
        os.remove(newOutput)

    def user_sample_rate(args_sample, inp):
        # (args_sample: int | str | None, inp) -> str | None:
        if(args_sample is None):
            if(len(inp.audio_streams) > 0):
                return inp.audio_streams[0]['samplerate']
            return None
        return str(args_sample)


    sampleRate = user_sample_rate(args.sample_rate, inp)
    log.debug('Samplerate: {}'.format(sampleRate))

    audioData = None
    audioExtensions = ['.wav', '.mp3', '.m4a', '.aiff', '.flac', '.ogg', '.oga',
        '.acc', '.nfa', '.mka']
    audioFile = inp.ext in audioExtensions
    if(audioFile):
        fps = 30 if args.force_fps_to is None else args.force_fps_to
        tracks = 1

        temp_file = os.path.join(TEMP, 'fastAud.wav')

        cmd = ['-i', inp.path]
        if(not fnone(args.audio_bitrate)):
            cmd.extend(['-b:a', args.audio_bitrate])
        cmd.extend(['-ac', '2', '-ar', sampleRate, '-vn', temp_file])
        ffmpeg.run(cmd)

        from auto_editor.scipy.wavfile import read
        sampleRate, audioData = read(temp_file)
    else:
        if(args.force_fps_to is not None):
            fps = args.force_fps_to
        else:
            fps = float(inp.fps)
            if(exporting_to_editor):
                fps = int(fps)

        if(fps < 1):
            log.error('{}: Frame rate cannot be below 1. fps: {}'.format(
                inp.basename, fps))

        tracks = len(inp.audio_streams)

        if(args.cut_by_this_track >= tracks and 'cut_by_this_track' in args._set):
            message = "You choose a track that doesn't exist.\nThere "
            if(tracks == 1):
                message += 'is only {} track.\n'.format(tracks)
            else:
                message += 'are only {} tracks.\n'.format(tracks)
            for t in range(tracks):
                message += ' Track {}\n'.format(t)
            log.error(message)

        def NumberOfVrfFrames(text, log):
            import re
            search = re.search(r'VFR:[\d.]+ \(\d+\/\d+\)', text, re.M)
            if(search is None):
                log.warning('Could not get number of VFR Frames.')
                return 0
            else:
                nums = re.search(r'\d+\/\d+', search.group()).group(0)
                log.debug(nums)
                return int(nums.split('/')[0])

        def hasVFR(cmd, log):
            return NumberOfVrfFrames(ffmpeg.pipe(cmd), log) != 0

        # Split audio tracks into: 0.wav, 1.wav, etc.
        cmd = ['-i', inp.path, '-hide_banner']
        for t in range(tracks):
            cmd.extend(['-map', '0:a:{}'.format(t)])
            if(not fnone(args.audio_bitrate)):
                cmd.extend(['-ab', args.audio_bitrate])
            cmd.extend(['-ac', '2', '-ar', sampleRate,
                os.path.join(TEMP, '{}.wav'.format(t))])
        cmd.extend(['-map', '0:v:0', '-vf', 'vfrdet', '-f', 'null', '-'])
        has_vfr = hasVFR(cmd, log)
        del cmd

        if(len(inp.video_streams) > 0 and tracks == 0):
            # Doesn't matter because we don't need to align to an audio track.
            has_vfr = False

        if(tracks != 0):
            if(args.cut_by_all_tracks):
                temp_file = os.path.join(TEMP, 'combined.wav')
                cmd = ['-i', inp.path, '-filter_complex',
                    '[0:a]amix=inputs={}:duration=longest'.format(tracks), '-ar',
                    sampleRate, '-ac', '2', '-f', 'wav', temp_file]
                ffmpeg.run(cmd)
                del cmd
            else:
                temp_file = os.path.join(TEMP, '{}.wav'.format(args.cut_by_this_track))

            from auto_editor.scipy.wavfile import read
            sampleRate, audioData = read(temp_file)

    log.debug('Frame Rate: {}'.format(fps))
    if(chunks is None):
        chunks = get_chunks(inp, speeds, segment, fps, args, log, audioData, sampleRate)
        effects = get_effects(audioData, sampleRate, fps, args, log)

    def isClip(chunk):
        return speeds[chunk[2]] != 99999

    def getNumberOfCuts(chunks, speeds):
        return len(list(filter(isClip, chunks)))

    def getClips(chunks, speeds):
        clips = []
        for chunk in chunks:
            if(isClip(chunk)):
                clips.append([chunk[0], chunk[1], speeds[chunk[2]] * 100])
        return clips

    num_cuts = getNumberOfCuts(chunks, speeds)
    clips = getClips(chunks, speeds)

    if(args.export_as_json):
        from auto_editor.formats.make_json import make_json_cutlist
        make_json_cutlist(inp.path, newOutput, auto_editor.version, chunks, speeds,
            log)
        return num_cuts, newOutput

    if(args.preview):
        from auto_editor.preview import preview
        preview(inp, chunks, speeds, log)
        return num_cuts, None

    if(args.export_to_premiere):
        from auto_editor.formats.premiere import premiere_xml
        premiere_xml(inp, TEMP, newOutput, clips, chunks, sampleRate, audioFile,
            fps, log)
        return num_cuts, newOutput

    if(args.export_to_final_cut_pro or args.export_to_resolve):
        from auto_editor.formats.final_cut_pro import fcp_xml

        totalFrames = chunks[len(chunks) - 1][1]
        fcp_xml(inp, TEMP, newOutput, clips, tracks, totalFrames, audioFile, fps, log)
        return num_cuts, newOutput

    if(args.export_to_shotcut):
        from auto_editor.formats.shotcut import shotcut_xml

        shotcut_xml(inp, TEMP, newOutput, clips, chunks, fps, log)
        return num_cuts, newOutput

    def makeAudioFile(inp, chunks, output):
        from auto_editor.render.audio import fastAudio, handleAudio, convertAudio
        theFile = handleAudio(ffmpeg, inp.path, args.audio_bitrate, str(sampleRate),
            TEMP, log)

        temp_file = os.path.join(TEMP, 'convert.wav')
        fastAudio(theFile, temp_file, chunks, speeds, log, fps,
            args.machine_readable_progress, args.no_progress)
        convertAudio(ffmpeg, temp_file, inp, output, args.audio_codec, log)

    if(audioFile):
        if(args.export_as_clip_sequence):
            i = 1
            for item in chunks:
                if(speeds[item[2]] == 99999):
                    continue
                makeAudioFile(inp, [item], append_filename(newOutput, '-{}'.format(i)))
                i += 1
        else:
            makeAudioFile(inp, chunks, newOutput)
        return num_cuts, newOutput

    def makeVideoFile(inp, chunks, output):
        from auto_editor.utils.video import handleAudioTracks, mux_rename_video
        continueVid = handleAudioTracks(ffmpeg, output, args, tracks, chunks,
            speeds, fps, TEMP, log)
        if(continueVid):
            if(args.render == 'auto'):
                if(args.zoom != [] or args.rectangle != []):
                    args.render = 'opencv'
                else:
                    try:
                        import av
                        args.render = 'av'
                    except ImportError:
                        args.render = 'opencv'

            log.debug('Using {} method'.format(args.render))
            if(args.render == 'av'):
                if(args.zoom != []):
                    log.error('Zoom effect is not supported on the '\
                        'av render method.')

                if(args.rectangle != []):
                    log.error('Rectangle effect is not supported on the '\
                        'av render method.')

                from auto_editor.render.av import render_av
                render_av(ffmpeg, inp, args, chunks, speeds, fps, has_vfr, TEMP, log)

            if(args.render == 'opencv'):
                from auto_editor.render.opencv import render_opencv
                render_opencv(ffmpeg, inp, args, chunks, speeds, fps, has_vfr,
                    effects, TEMP, log)

            if(log.is_debug):
                log.debug('Writing the output file.')
            else:
                log.conwrite('Writing the output file.')

            mux_rename_video(ffmpeg, output, args, tracks, TEMP, log)
            if(output is not None and not os.path.isfile(output)):
                log.bug('The file {} was not created.'.format(output))

    if(args.export_as_clip_sequence):

        def pad_chunk(item, totalFrames):
            start = None
            end = None
            if(item[0] != 0):
                start = [0, item[0], 2]
            if(item[1] != totalFrames -1):
                end = [item[1], totalFrames -1, 2]

            if(start is None):
                return [item] + [end]
            if(end is None):
                return [start] + [item]
            return [start] + [item] + [end]

        i = 1
        totalFrames = chunks[len(chunks) - 1][1]
        speeds.append(99999) # guarantee we have a cut speed to work with.
        for chunk in chunks:
            if(speeds[chunk[2]] == 99999):
                continue

            makeVideoFile(inp, pad_chunk(chunk, totalFrames),
                append_filename(newOutput, '-{}'.format(i)))
            i += 1
    else:
        makeVideoFile(inp, chunks, newOutput)
    return num_cuts, newOutput


def main():
    dir_path = os.path.dirname(os.path.realpath(__file__))

    parser = vanparse.ArgumentParser('Auto-Editor', auto_editor.version,
        description='\nAuto-Editor is an automatic video/audio creator and editor. '\
            'By default, it will detect silence and create a new video with those '\
            'sections cut out. By changing some of the options, you can export to a '\
            'traditional editor like Premiere Pro and adjust the edits there, adjust '\
            'the pacing of the cuts, and change the method of editing like using audio '\
            'loudness and video motion to judge making cuts.\nRun:\n    auto-editor '\
            '--help\n\nTo get the list of options.\n')

    subcommands = ['create', 'test', 'info', 'levels', 'grep']

    if(len(sys.argv) > 1 and sys.argv[1] in subcommands):
        if(sys.argv[1] == 'create'):
            from auto_editor.subcommands.create import create as sub
        if(sys.argv[1] == 'test'):
            from auto_editor.subcommands.test import test as sub
        if(sys.argv[1] == 'info'):
            from auto_editor.subcommands.info import info as sub
        if(sys.argv[1] == 'levels'):
            from auto_editor.subcommands.levels import levels as sub
        if(sys.argv[1] == 'grep'):
            from auto_editor.subcommands.grep import grep as sub

        sub(sys.argv[2:])

        sys.exit()
    else:
        parser = main_options(parser)
        args = parser.parse_args(sys.argv[1:], Log(), 'auto-editor')

    timer = Timer(args.quiet)

    exporting_to_editor = (args.export_to_premiere or args.export_to_resolve or
        args.export_to_final_cut_pro or args.export_to_shotcut)
    making_data_file = exporting_to_editor or args.export_as_json

    is64bit = '64-bit' if sys.maxsize > 2**32 else '32-bit'

    if(args.debug and args.input == []):
        import platform
        log = Log()
        ffmpeg = FFmpeg(dir_path, args.ffmpeg_location, args.my_ffmpeg,
            args.show_ffmpeg_debug, log)

        print('Python Version: {} {}'.format(platform.python_version(), is64bit))
        print('Platform: {} {}'.format(platform.system(), platform.release()))
        print('Config File path: {}'.format(os.path.join(dir_path, 'config.txt')))
        print('FFmpeg path: {}'.format(ffmpeg.getPath()))
        print('FFmpeg version: {}'.format(ffmpeg.getVersion()))
        print('Auto-Editor version {}'.format(auto_editor.version))
        sys.exit()

    if(is64bit == '32-bit'):
        log.warning('You have the 32-bit version of Python, which may lead to '\
            'memory crashes.')

    if(args.version):
        print('Auto-Editor version {}'.format(auto_editor.version))
        sys.exit()

    TEMP = tempfile.mkdtemp()
    log = Log(args.debug, args.quiet, temp=TEMP)

    ffmpeg = FFmpeg(dir_path, args.ffmpeg_location, args.my_ffmpeg,
        args.show_ffmpeg_debug, log)

    log.debug('Temp Directory: {}'.format(TEMP))

    if(args.input == []):
        log.error('You need to give auto-editor an input file or folder so it can ' \
            'do the work for you.')

    if([args.export_to_premiere, args.export_to_resolve,
        args.export_to_final_cut_pro, args.export_as_audio,
        args.export_to_shotcut, args.export_as_clip_sequence].count(True) > 1):
        log.error('You must choose only one export option.')

    if(making_data_file and (args.video_codec != 'uncompressed' or
        args.constant_rate_factor != 'unset' or args.tune != 'unset')):
        log.warning('exportMediaOps options are not used when making a data file.')

    if(isinstance(args.frame_margin, str)):
        try:
            if(float(args.frame_margin) < 0):
                log.error('Frame margin cannot be negative.')
        except ValueError:
            log.error('Frame margin {}, is not valid.'.format(args.frame_margin))
    elif(args.frame_margin < 0):
        log.error('Frame margin cannot be negative.')
    if(args.constant_rate_factor != 'unset'):
        if(int(args.constant_rate_factor) < 0 or int(args.constant_rate_factor) > 51):
            log.error('Constant rate factor (crf) must be between 0-51.')
    if(args.width < 1):
        log.error('motionOps --width cannot be less than 1.')
    if(args.dilates < 0):
        log.error('motionOps --dilates cannot be less than 0')


    def write_starting_message(args, log):
        if(args.preview):
            pass
        elif(args.export_to_premiere):
            log.conwrite('Exporting to Adobe Premiere Pro XML file.')
        elif(args.export_to_final_cut_pro):
            log.conwrite('Exporting to Final Cut Pro XML file.')
        elif(args.export_to_resolve):
            log.conwrite('Exporting to DaVinci Resolve XML file.')
        elif(args.export_to_shotcut):
            log.conwrite('Exporting to Shotcut XML Timeline file.')
        elif(args.export_as_audio):
            log.conwrite('Exporting as audio.')
        else:
            log.conwrite('Starting.')

    write_starting_message(args, log)

    if(args.preview or args.export_as_clip_sequence or making_data_file):
        args.no_open = True

    args.background = usefulfunctions.hex_to_bgr(args.background, log)
    if(args.blur < 0):
        args.blur = 0

    if(args.silent_speed <= 0 or args.silent_speed > 99999):
        args.silent_speed = 99999

    if(args.video_speed <= 0 or args.video_speed > 99999):
        args.video_speed = 99999

    if(args.output_file is None):
        args.output_file = []

    from auto_editor.validateInput import valid_input
    inputList, segments = valid_input(args.input, ffmpeg, args, log)

    if(len(args.output_file) < len(inputList)):
        for i in range(len(inputList) - len(args.output_file)):
            args.output_file.append(set_output_name(inputList[i], making_data_file, args))

    if(args.combine_files):
        temp_file = os.path.join(TEMP, 'combined.mp4')
        cmd = []
        for fileref in inputList:
            cmd.extend(['-i', fileref])
        cmd.extend(['-filter_complex', '[0:v]concat=n={}:v=1:a=1'.format(len(inputList)),
            '-codec:v', 'h264', '-pix_fmt', 'yuv420p', '-strict', '-2', temp_file])
        ffmpeg.run(cmd)
        del cmd
        inputList = [temp_file]

    speeds = [args.silent_speed, args.video_speed]
    if(args.cut_out != [] and 99999 not in speeds):
        speeds.append(99999)

    for item in args.set_speed_for_range:
        if(item[0] not in speeds):
            speeds.append(float(item[0]))

    log.debug('Speeds: {}'.format(speeds))

    def main_loop(inputList, ffmpeg, args, speeds, segments, log):
        num_cuts = 0

        for i, INPUT_FILE in enumerate(inputList):
            inp = ffmpeg.file_info(INPUT_FILE)

            if(len(inputList) > 1):
                log.conwrite('Working on {}'.format(inp.basename))

            cuts, newOutput = edit_media(i, inp, ffmpeg, args, speeds, segments[i],
                exporting_to_editor, making_data_file, TEMP, log)
            num_cuts += cuts

        if(not args.preview and not making_data_file):
            timer.stop()

        if(not args.preview and making_data_file):
            # Assume making each cut takes about 30 seconds.
            time_save = usefulfunctions.human_readable_time(num_cuts * 30)
            s = 's' if num_cuts != 1 else ''

            log.print('Auto-Editor made {} cut{}, which would have taken about {} if '\
                'edited manually.'.format(num_cuts, s, time_save))

        if(not args.no_open):
            usefulfunctions.open_with_system_default(newOutput, log)

    try:
        main_loop(inputList, ffmpeg, args, speeds, segments, log)
    except KeyboardInterrupt:
        log.error('Keyboard Interrupt')
    log.cleanup()

if(__name__ == '__main__'):
    main()



'''subcommands/info.py'''

from __future__ import print_function

def info_options(parser):
    parser.add_argument('--include_vfr', action='store_true',
        help='skip information that is very slow to get.')
    parser.add_argument('--ffmpeg_location', default=None,
        help='point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    parser.add_argument('(input)', nargs='*',
        help='the path to a file you want inspected.')
    return parser

def info(sys_args=None):
    import os
    import sys

    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.func import clean_list, aspect_ratio
    from auto_editor.utils.log import Log

    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('info', auto_editor.version,
        description='Get basic information about media files.')
    parser = info_options(parser)

    if(sys_args is None):
        sys_args = sys.args[1:]

    log = Log()
    args = parser.parse_args(sys_args, log, 'info')

    dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    ffmpeg = FFmpeg(dir_path, args.ffmpeg_location, args.my_ffmpeg, False, log)

    def aspect_str(w, h):
        w, h = aspect_ratio(int(w), int(h))
        if(w is None):
            return ''
        return ' ({}:{})'.format(w, h)

    for file in args.input:
        text = ''
        if(os.path.exists(file)):
            text += 'file: {}\n'.format(file)
        else:
            log.error('Could not find file: {}'.format(file))

        inp = ffmpeg.file_info(file)

        if(len(inp.video_streams) > 0):
            text += ' - fps: {}\n'.format(inp.fps)
            text += ' - duration: {}\n'.format(inp.duration)

            w = inp.video_streams[0]['width']
            h = inp.video_streams[0]['height']

            if(w is not None and h is not None):
                text += ' - resolution: {}x{}{}\n'.format(w, h, aspect_str(w, h))

            text += ' - video codec: {}\n'.format(inp.video_streams[0]['codec'])
            text += ' - video bitrate: {}\n'.format(inp.video_streams[0]['bitrate'])

            tracks = len(inp.audio_streams)
            text += ' - audio tracks: {}\n'.format(tracks)

            for track in range(tracks):
                text += '   - Track #{}\n'.format(track)
                text += '     - codec: {}\n'.format(inp.audio_streams[track]['codec'])
                text += '     - samplerate: {}\n'.format(
                    inp.audio_streams[track]['samplerate'])
                text += '     - bitrate: {}\n'.format(inp.audio_streams[track]['bitrate'])

            sub_tracks = len(inp.subtitle_streams)
            if(sub_tracks > 0):
                text += ' - subtitle tracks: {}\n'.format(sub_tracks)
                for track in range(tracks):
                    text += '   - Track #{}\n'.format(track)
                    text += '     - lang: {}\n'.format(inp.subtitle_streams[track]['lang'])

            if(args.include_vfr):
                print(text, end='')
                text = ''
                fps_mode = ffmpeg.pipe(['-i', file, '-hide_banner', '-vf', 'vfrdet',
                    '-an', '-f', 'null', '-'])
                fps_mode = clean_list(fps_mode.split('\n'), '\r\t')
                fps_mode = fps_mode.pop()

                if('VFR:' in fps_mode):
                    fps_mode = (fps_mode[fps_mode.index('VFR:'):]).strip()

                text += ' - {}\n'.format(fps_mode)

        elif(len(inp.audio_streams) > 0):
            text += ' - duration: {}\n'.format(inp.duration)
            text += ' - codec: {}\n'.format(inp.audio_streams[0]['codec'])
            text += ' - samplerate: {}\n'.format(inp.audio_streams[0]['samplerate'])
            text += ' - bitrate: {}\n'.format(inp.audio_streams[0]['bitrate'])
        else:
            text += 'Invalid media.\n'
        print(text)

if(__name__ == '__main__'):
    info()



'''preview.py'''

# Internal Libraries
from datetime import timedelta

# Included Libraries
from auto_editor.utils.func import get_new_length

def printTimeFrame(title, frames, fps):
    in_sec = round(frames / fps, 1)
    fps = round(fps)
    if(in_sec < 1):
        minutes = '{}/{} frames'.format(int(frames), fps)
    else:
        minutes = timedelta(seconds=round(in_sec))
    print('{}: {} secs ({})'.format(title, in_sec, minutes))


def preview(inp, chunks, speeds, log):
    fps = 30 if inp.fps is None else float(inp.fps)

    old_time = chunks[len(chunks)-1][1]
    print('')
    printTimeFrame('Old length', old_time, fps)

    new_length = get_new_length(chunks, speeds, fps)
    printTimeFrame('New length', new_length * fps, fps)
    print('')

    clips = 0
    cuts = 0
    cut_lens = []
    clip_lens = []
    for chunk in chunks:
        state = chunk[2]
        if(speeds[state] != 99999):
            clips += 1
            leng = (chunk[1] - chunk[0]) / speeds[state]
            clip_lens.append(leng)
        else:
            cuts += 1
            leng = chunk[1] - chunk[0]
            cut_lens.append(leng)

    print('Number of clips: {}'.format(clips))
    printTimeFrame('Smallest clip length', min(clip_lens), fps)
    printTimeFrame('Largest clip length', max(clip_lens), fps)
    printTimeFrame('Average clip length', sum(clip_lens) / len(clip_lens), fps)
    print('\nNumber of cuts: {}'.format(cuts))

    if(cut_lens != []):
        printTimeFrame('Smallest cut length', min(cut_lens), fps)
        printTimeFrame('Largest cut length', max(cut_lens), fps)
        printTimeFrame('Average cut length', sum(cut_lens) / len(cut_lens), fps)
        print('')

    log.debug('Chunks:\n{}'.format(chunks))



'''utils/func.py'''

from __future__ import division

"""
To prevent duplicate code being pasted between scripts, common functions should be
put here. No function should modify or create video/audio files on its own.
"""

def get_stdout(cmd):
    import subprocess
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT)
    stdout, __ = process.communicate()
    return stdout.decode()

def clean_list(x, rm_chars):
    # (x: list, rm_chars: str) -> list
    new_list = []
    for item in x:
        for char in rm_chars:
            item = item.replace(char, '')
        new_list.append(item)
    return new_list

def aspect_ratio(width, height):
    # (width: int, height: int) -> tuple(2)
    if(height == 0):
        return None, None

    def gcd(a, b):
        while b:
            a, b = b, a % b
        return a

    c = gcd(width, height)
    return width // c, height // c

def get_new_length(chunks, speeds, fps):
    # (chunks: list, speeds: list, fps: float) -> float
    timeInFrames = 0
    for chunk in chunks:
        leng = chunk[1] - chunk[0]
        if(speeds[chunk[2]] < 99999):
            timeInFrames += leng * (1 / speeds[chunk[2]])
    return timeInFrames / fps

def human_readable_time(time_in_secs):
    units = 'seconds'
    if(time_in_secs >= 3600):
        time_in_secs = round(time_in_secs / 3600, 1)
        if(time_in_secs % 1 == 0):
            time_in_secs = round(time_in_secs)
        units = 'hours'
    if(time_in_secs >= 60):
        time_in_secs = round(time_in_secs / 60, 1)
        if(time_in_secs >= 10 or time_in_secs % 1 == 0):
            time_in_secs = round(time_in_secs)
        units = 'minutes'
    return '{} {}'.format(time_in_secs, units)

def open_with_system_default(path, log):
    from subprocess import call
    try:  # should work on Windows
        from os import startfile
        startfile(path)
    except ImportError:
        try:  # should work on MacOS and most Linux versions
            call(['open', path])
        except Exception:
            try: # should work on WSL2
                call(['cmd.exe', '/C', 'start', path])
            except Exception:
                try: # should work on various other Linux distros
                    call(['xdg-open', path])
                except Exception:
                    log.warning('Could not open output file.')

def hex_to_bgr(hex_str, log):
    import re
    if(re.compile(r'#[a-fA-F0-9]{3}(?:[a-fA-F0-9]{3})?$').match(hex_str)):
        if(len(hex_str) < 5):
            return [int(hex_str[i]*2, 16) for i in (3, 2, 1)]
        return [int(hex_str[i:i+2], 16) for i in (5, 3, 1)]
    log.error('Invalid hex code: {}'.format(hex_str))

def fnone(val):
    return val == 'none' or val == 'unset' or val is None

def append_filename(name, val):
    dot_index = name.rfind('.')
    return name[:dot_index] + val + name[dot_index:]

def term_size():
    try:
        from shutil import get_terminal_size
        return get_terminal_size()
    except ImportError:
        class Terminal:
            def __init__(self):
                self.columns = 80
        return Terminal()



'''subcommands/test.py'''

from __future__ import print_function

"""
Test auto-editor and make sure everything is working.
"""

# Internal Libraries
import os
import sys
import shutil
import platform
import subprocess

# Included Libraries
from auto_editor.utils.func import clean_list
from auto_editor.utils.log import Log
import auto_editor.vanparse as vanparse

def test_options(parser):
    parser.add_argument('--ffprobe_location', default='ffprobe',
        help='point to your custom ffmpeg file.')
    parser.add_argument('--only', '-n', nargs='*')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    return parser

class FFprobe():
    def __init__(self, path):
        self.path = path

    def run(self, cmd):
        cmd.insert(0, self.path)
        process = subprocess.Popen(cmd, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        stdout, __ = process.communicate()
        return stdout.decode('utf-8')

    def pipe(self, cmd):
        full_cmd = [self.path, '-v', 'error'] + cmd
        process = subprocess.Popen(full_cmd, stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT)
        stdout, __ = process.communicate()
        return stdout.decode('utf-8')

    def _get(self, file, stream, the_type, track, of='compact=p=0:nk=1'):
        return self.pipe(['-select_streams', '{}:{}'.format(the_type, track),
            '-show_entries', 'stream={}'.format(stream), '-of', of, file]).strip()

    def getResolution(self, file):
        return self._get(file, 'height,width', 'v', 0, of='csv=s=x:p=0')

    def getTimeBase(self, file):
        return self.pipe(['-select_streams', 'v', '-show_entries',
            'stream=avg_frame_rate', '-of', 'compact=p=0:nk=1', file]).strip()

    def getFrameRate(self, file):
        nums = clean_list(self.getTimeBase(file).split('/'), '\r\t\n')
        return int(nums[0]) / int(nums[1])

    def getAudioCodec(self, file, track=0):
        return self._get(file, 'codec_name', 'a', track)

    def getVideoCodec(self, file, track=0):
        return self._get(file, 'codec_name', 'v', track)

    def getSampleRate(self, file, track=0):
        return self._get(file, 'sample_rate', 'a', track)


def pipe_to_console(cmd):
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stderr = process.communicate()
    return process.returncode, stdout.decode('utf-8'), stderr.decode('utf-8')

def cleanup(the_dir):
    for item in os.listdir(the_dir):
        item = os.path.join(the_dir, item)
        if('_ALTERED' in item or item.endswith('.xml') or item.endswith('.json')
            or item.endswith('.fcpxml') or item.endswith('.mlt')):
            os.remove(item)
        if(item.endswith('_tracks')):
            shutil.rmtree(item)

def clean_all():
    cleanup('resources')
    cleanup(os.getcwd())


def getRunner():
    if(platform.system() == 'Windows'):
        return ['py', '-m', 'auto_editor']
    return ['python3', '-m', 'auto_editor']


def run_program(cmd):
    cmd = getRunner() + cmd
    if('.' in cmd[0]):
        cmd += ['--no_open']

    returncode, stdout, stderr = pipe_to_console(cmd)
    if(returncode > 0):
        raise Exception('Test Failed.\n{}\n{}\n'.format(stdout, stderr))


def checkForError(cmd, match=None):
    returncode, stdout, stderr = pipe_to_console(getRunner() + cmd)
    if(returncode > 0):
        if('Error!' in stderr):
            if(match is not None and match not in stderr):
                raise Exception('Test Failed.\nCould\'t find "{}"'.format(match))
        else:
            raise Exception('Test Failed.\nProgram crashed.\n{}\n{}'.format(stdout, stderr))
    else:
        raise Exception('Test Failed.\nProgram should not responsed with a code 0.')

def fullInspect(fileName, *args):
    for item in args:
        func = item[0]
        expectedOutput = item[1]
        if(func(fileName) != expectedOutput):
            # Cheating on float numbers to allow 30 to equal 29.99944409236961
            if(isinstance(expectedOutput, float)):
                from math import ceil
                if(ceil(func(fileName) * 100) == expectedOutput * 100):
                    continue
            print('Expected Value: {} {}'.format(expectedOutput, type(expectedOutput)))
            print('Actual Value: {} {}'.format(func(fileName), type(func(fileName))))
            raise Exception('Inspection Failed.')


class Tester():
    def __init__(self, args):
        self.passed_tests = 0
        self.failed_tests = 0
        self.allowable_fails = 0
        self.args = args

    def run_test(self, name, func, description='', cleanup=None):
        if(self.args.only != [] and name not in self.args.only):
            return
        try:
            func()
        except Exception as e:
            self.failed_tests += 1
            print('{} Failed.'.format(name))
            print(e)
            clean_all()
            if(self.failed_tests > self.allowable_fails):
                sys.exit(1)
        else:
            self.passed_tests += 1
            print('{} Passed.'.format(name))
            if(cleanup is not None):
                cleanup()

    def end(self):
        print('{}/{}'.format(self.passed_tests, self.passed_tests + self.failed_tests))
        clean_all()
        if(self.failed_tests > self.allowable_fails):
            sys.exit(1)
        sys.exit(0)

def test(sys_args=None):
    parser = vanparse.ArgumentParser('test', 'version')
    parser = test_options(parser)

    if(sys_args is None):
        sys_args = sys.args[1:]

    args = parser.parse_args(sys_args, Log(), 'test')
    ffprobe = FFprobe(args.ffprobe_location)

    tester = Tester(args)

    def help_tests():
        run_program(['--help'])
        run_program(['-h'])
        run_program(['--frame_margin', '--help'])
        run_program(['--frame_margin', '-h'])
        run_program(['exportMediaOps', '--help'])
        run_program(['exportMediaOps', '-h'])
        run_program(['progressOps', '-h'])

        run_program(['--help', '--help'])
        run_program(['-h', '--help'])
        run_program(['--help', '-h'])
        run_program(['-h', '--help'])
    tester.run_test('help_tests', help_tests, description='check the help option, '\
        'its short, and help on options and groups.')

    def version_debug():
        run_program(['--version'])
        run_program(['-v'])
        run_program(['-V'])

        run_program(['--debug'])

        # sanity check for example.mp4/ffprobe
        if(ffprobe.getFrameRate('example.mp4') != 30.0):
            print('getFrameRate did not equal 30.0')
            sys.exit(1)
    tester.run_test('version_tests', version_debug)

    def info_tests():
        run_program(['info', 'example.mp4'])
        run_program(['info', 'resources/man_on_green_screen.mp4'])
        run_program(['info', 'resources/multi-track.mov'])
        run_program(['info', 'resources/newCommentary.mp3'])
        run_program(['info', 'resources/test.mkv'])
    tester.run_test('info_tests', info_tests)

    def level_tests():
        run_program(['levels', 'example.mp4'])
        run_program(['levels', 'resources/newCommentary.mp3'])
    tester.run_test('level_tests', level_tests, lambda a: os.remove('data.txt'))

    def example_tests():
        run_program(['example.mp4'])
        fullInspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1280x720'],
            [ffprobe.getSampleRate, '48000'],
        )
        run_program(['example.mp4', '--video_codec', 'uncompressed'])
        fullInspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1280x720'],
            [ffprobe.getVideoCodec, 'mpeg4'],
            [ffprobe.getSampleRate, '48000'],
        )
    tester.run_test('example_tests', example_tests)

    def gif_test():
        run_program(['resources/man_on_green_screen.gif', '--edit', 'none'])
    tester.run_test('gif_test', gif_test, description='run gif files', cleanup=clean_all)

    def render_tests():
        run_program(['example.mp4', '--render', 'opencv'])
        fullInspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1280x720'],
            [ffprobe.getSampleRate, '48000'],
        )
    tester.run_test('render_tests', render_tests)

    def margin_tests():
        run_program(['example.mp4', '-m', '3'])
        run_program(['example.mp4', '--margin', '3'])
        run_program(['example.mp4', '-m', '0.3sec'])
    tester.run_test('margin_tests', margin_tests)

    def extension_tests():
        shutil.copy('example.mp4', 'example')
        checkForError(['example', '--no_open'], 'must have an extension.')
        os.remove('example')

        run_program(['example.mp4', '-o', 'example.mkv'])
        os.remove('example.mkv')

        run_program(['resources/test.mkv', '-o', 'test.mp4'])
        os.remove('test.mp4')

    tester.run_test('extension_tests', extension_tests)

    def progress_ops_test():
        run_program(['example.mp4', 'progressOps', '--machine_readable_progress'])
        run_program(['example.mp4', 'progressOps', '--no_progress'])
    tester.run_test('progress_ops_test', progress_ops_test)

    def silent_threshold():
        run_program(['resources/newCommentary.mp3', '--silent_threshold', '0.1'])
    tester.run_test('silent_threshold', silent_threshold)

    def track_tests():
        run_program(['resources/multi-track.mov', '--cut_by_all_tracks'])
        run_program(['resources/multi-track.mov', '--keep_tracks_seperate'])
        run_program(['example.mp4', '--cut_by_this_audio', 'resources/newCommentary.mp3'])
    tester.run_test('track_tests', track_tests)

    def json_tests():
        run_program(['example.mp4', '--export_as_json'])
        run_program(['example.json'])
    tester.run_test('json_tests', json_tests)

    def speed_tests():
        run_program(['example.mp4', '-s', '2', '-mcut', '10'])
        run_program(['example.mp4', '-v', '2', '-mclip', '4'])
        run_program(['example.mp4', '--sounded_speed', '0.5'])
        run_program(['example.mp4', '--silent_speed', '0.5'])
    tester.run_test('speed_tests', speed_tests)

    def scale_tests():
        run_program(['example.mp4', '--scale', '1.5', '--render', 'av'])
        fullInspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1920x1080'],
            [ffprobe.getSampleRate, '48000'],
        )

        run_program(['example.mp4', '--scale', '0.2', '--render', 'av'])
        fullInspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '256x144'],
            [ffprobe.getSampleRate, '48000'],
        )

        run_program(['example.mp4', '--scale', '1.5', '--render', 'opencv'])
        fullInspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '1920x1080'],
            [ffprobe.getSampleRate, '48000'],
        )

        run_program(['example.mp4', '--scale', '0.2', '--render', 'opencv'])
        fullInspect(
            'example_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '256x144'],
            [ffprobe.getSampleRate, '48000'],
        )
    tester.run_test('scale_tests', scale_tests)

    def various_errors_test():
        checkForError(['example.mp4', '--zoom', '0,60,1.5', '--render', 'av'])
        checkForError(['example.mp4', '--zoom', '0'])
        checkForError(['example.mp4', '--zoom', '0,60'])
        checkForError(['example.mp4', '--rectangle', '0,60,0,10,10,20', '--render', 'av'])
        checkForError(['example.mp4', '--rectangle', '0,60'])
        checkForError(['example.mp4', '--background', '000'])
    tester.run_test('various_errors_test', various_errors_test)

    def create_sub_test():
        run_program(['create', 'test', '--width', '640', '--height', '360', '-o',
            'testsrc.mp4'])
        fullInspect(
            'testsrc.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '640x360'],
        )
    tester.run_test('create_sub_test', create_sub_test)

    def effect_tests():
        run_program(['testsrc.mp4', '--mark_as_loud', 'start,end', '--zoom', '10,60,2'])

        run_program(['example.mp4', '--mark_as_loud', 'start,end', '--rectangle',
            'audio>0.05,audio<0.05,20,50,50,100', 'audio>0.1,audio<0.1,120,50,150,100'])

        run_program(['testsrc.mp4', '--mark_as_loud', 'start,end', '--zoom',
            'start,end,1,0.5,centerX,centerY,linear', '--scale', '0.5'])
        fullInspect(
            'testsrc_ALTERED.mp4',
            [ffprobe.getFrameRate, 30.0],
            [ffprobe.getResolution, '320x180'],
        )
        run_program(['testsrc.mp4', '--mark_as_loud', 'start,end', '--rectangle',
            '0,30,0,200,100,300,#43FA56,10'])
        os.remove('testsrc_ALTERED.mp4')
        os.remove('testsrc.mp4')
    tester.run_test('effect_tests', effect_tests,
        description='test the zoom and rectangle options',
        cleanup=clean_all)

    def export_tests():
        for item in os.listdir('resources'):
            if('man_on_green_screen' in item or item.startswith('.')):
                continue
            item = 'resources/{}'.format(item)
            run_program([item])
            run_program([item, '-exp'])
            run_program([item, '-exr'])
            run_program([item, '-exf'])
            run_program([item, '-exs'])
            run_program([item, '--export_as_clip_sequence'])
            run_program([item, '--preview'])
            cleanup('resources')
    tester.run_test('export_tests', export_tests)

    def codec_tests():
        run_program(['example.mp4', '--video_codec', 'h264', '--preset', 'faster'])
        run_program(['example.mp4', '--audio_codec', 'ac3'])
        run_program(['resources/newCommentary.mp3', 'exportMediaOps', '-acodec', 'pcm_s16le'])
    tester.run_test('codec_tests', codec_tests)

    def combine_tests():
        run_program(['example.mp4', '--mark_as_silent', '0,171', '-o', 'hmm.mp4'])
        run_program(['example.mp4', 'hmm.mp4', '--combine_files', '--debug'])
        os.remove('hmm.mp4')
    tester.run_test('combine_tests', combine_tests)

    def motion_tests():
        run_program(['resources/man_on_green_screen.mp4', '--edit_based_on', 'motion',
         '--debug', '--frame_margin', '0', '-mcut', '0', '-mclip', '0'])
        run_program(['resources/man_on_green_screen.mp4', '--edit_based_on', 'motion',
            '--motion_threshold', '0'])
    tester.run_test('motion_tests', motion_tests)

    tester.end()

if(__name__ == '__main__'):
    test()



'''render/av.py'''

# Internal Libraries
import os.path
import subprocess

# Included Libaries
from auto_editor.utils.progressbar import ProgressBar
from .utils import properties, scale_to_sped

def render_av(ffmpeg, inp, args, chunks, speeds, fps, has_vfr, temp, log):
    import av

    totalFrames = chunks[len(chunks) - 1][1]
    videoProgress = ProgressBar(totalFrames, 'Creating new video',
        args.machine_readable_progress, args.no_progress)

    if(has_vfr):
        class Wrapper:
            """
            Wrapper which only exposes the `read` method to avoid PyAV
            trying to use `seek`.
            From: github.com/PyAV-Org/PyAV/issues/578#issuecomment-621362337
            """

            name = "<wrapped>"

            def __init__(self, fh):
                self._fh = fh

            def read(self, buf_size):
                return self._fh.read(buf_size)

        # Create a cfr stream on stdout.
        cmd = ['-i', inp.path, '-map', '0:v:0', '-vf', 'fps=fps={}'.format(fps), '-r',
            str(fps), '-vsync', '1', '-f', 'matroska', '-vcodec', 'rawvideo', 'pipe:1']

        wrapper = Wrapper(ffmpeg.Popen(cmd).stdout)
        input_ = av.open(wrapper, 'r')
    else:
        input_ = av.open(inp.path)

    inputVideoStream = input_.streams.video[0]
    inputVideoStream.thread_type = 'AUTO'

    width = inputVideoStream.width
    height = inputVideoStream.height
    pix_fmt = inputVideoStream.pix_fmt

    log.debug('pix_fmt: {}'.format(pix_fmt))

    cmd = [ffmpeg.getPath(), '-hide_banner', '-y', '-f', 'rawvideo', '-vcodec', 'rawvideo',
        '-pix_fmt', pix_fmt, '-s', '{}*{}'.format(width, height), '-framerate', str(fps),
        '-i', '-', '-pix_fmt', pix_fmt]

    if(args.scale != 1):
        cmd.extend(['-vf', 'scale=iw*{}:ih*{}'.format(args.scale, args.scale),
            os.path.join(temp, 'scale.mp4')])
    else:
        cmd = properties(cmd, args, inp)
        cmd.append(os.path.join(temp, 'spedup.mp4'))

    if(args.show_ffmpeg_debug):
        process2 = subprocess.Popen(cmd, stdin=subprocess.PIPE)
    else:
        process2 = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL)

    inputEquavalent = 0.0
    outputEquavalent = 0
    index = 0
    chunk = chunks.pop(0)

    try:
        for packet in input_.demux(inputVideoStream):
            for frame in packet.decode():
                index += 1
                if(len(chunks) > 0 and index >= chunk[1]):
                    chunk = chunks.pop(0)

                if(speeds[chunk[2]] != 99999):
                    inputEquavalent += (1 / speeds[chunk[2]])

                while inputEquavalent > outputEquavalent:
                    in_bytes = frame.to_ndarray().tobytes()
                    process2.stdin.write(in_bytes)
                    outputEquavalent += 1

                videoProgress.tick(index - 1)
        process2.stdin.close()
        process2.wait()
    except BrokenPipeError:
        log.print(cmd)
        process2 = subprocess.Popen(cmd, stdin=subprocess.PIPE)
        log.error('Broken Pipe Error!')

    if(args.scale != 1):
        scale_to_sped(ffmpeg, inp, args, temp)




'''cutting.py'''

from __future__ import print_function

import numpy as np

def combine_audio_motion(audioList, motionList, based, log):
    # (audioList: np.ndarray, motionList: np.ndarray, based: str, log) -> np.ndarray:

    if(based == 'audio' or based == 'not_audio'):
        if(max(audioList) == 0):
            log.error('There was no place where audio exceeded the threshold.')
    if(based == 'motion' or based == 'not_motion'):
        if(max(motionList) == 0):
            log.error('There was no place where motion exceeded the threshold.')

    if(audioList is not None and max(audioList) == 0):
        log.warning('There was no place where audio exceeded the threshold.')
    if(motionList is not None and max(motionList) == 0):
        log.warning('There was no place where motion exceeded the threshold.')

    if(based == 'audio'):
        return audioList

    if(based == 'motion'):
        return motionList

    if(based == 'not_audio'):
        return np.invert(audioList)

    if(based == 'not_motion'):
        return np.invert(motionList)

    if(based == 'audio_and_motion'):
        return audioList & motionList

    if(based == 'audio_or_motion'):
        return audioList | motionList

    if(based == 'audio_xor_motion'):
        return np.bitwise_xor(audioList, motionList)

    if(based == 'audio_and_not_motion'):
        return audioList & np.invert(motionList)

    if(based == 'not_audio_and_motion'):
        return np.invert(audioList) & motionList

    if(based == 'not_audio_and_not_motion'):
        return np.invert(audioList) & np.invert(motionList)
    return None


def combine_segment(has_loud, segment, fps):
    for item in segment:
        start, end = item['segment']
        start = int(start * fps)
        end = int(end * fps)
        has_loud[start:end] = False
    return has_loud


def removeSmall(has_loud, lim, replace, with_):
    # (has_loud: np.ndarray, lim: int, replace: bool, with_: bool) -> np.ndarray:
    startP = 0
    active = False
    for j, item in enumerate(has_loud):
        if(item == replace):
            if(not active):
                startP = j
                active = True
            # Special case for end.
            if(j == len(has_loud) - 1):
                if(j - startP < lim):
                    has_loud[startP:j+1] = with_
        else:
            if(active):
                if(j - startP < lim):
                    has_loud[startP:j] = with_
                active = False
    return has_loud


def is_number(val):
    return val.replace('.', '', 1).isdigit()

def setRange(has_loud, syntaxRange, fps, with_, log):
    # (has_loud: np.ndarray, syntaxRange, fps: float, with_, log) -> np.ndarray:

    def replace_variables_to_values(item):
        if(is_number(item)):
            return int(item)
        if(item == 'start'):
            return 0
        if(item == 'end'):
            return len(has_loud)
        if(item.startswith('sec')):
            log.error('Seconds unit not implemented in this function.')
        log.error('Variable {} not avaiable in this context.'.format(item))

    def var_val_to_frames(num):
        # (num) -> int:
        num = replace_variables_to_values(num)
        if(num < 0):
            num = len(has_loud) - num
        return num

    for item in syntaxRange:
        pair = list(map(var_val_to_frames, item))
        has_loud[pair[0]:pair[1]] = with_
    return has_loud


def seconds_to_frames(value, fps):
    # (int | str, float) -> int:
    if(isinstance(value, str)):
        return int(float(value) * fps)
    return value

def cook(has_loud, minClip, minCut):
    # (has_loud: np.ndarray, minClip: int, minCut: int) -> np.ndarray:
    has_loud = removeSmall(has_loud, minClip, replace=1, with_=0)
    has_loud = removeSmall(has_loud, minCut, replace=0, with_=1)
    return has_loud


# Turn long silent/loud array to formatted chunk list.
# Example: [True, True, True, False, False] => [[0, 3, 1], [3, 5, 0]]
def chunkify(has_loud, has_loud_length=None):
    # (np.ndarray, int) -> list:
    if(has_loud_length is None):
        has_loud_length = len(has_loud)

    chunks = []
    startP = 0
    for j in range(1, has_loud_length):
        if(has_loud[j] != has_loud[j - 1]):
            chunks.append([startP, j, int(has_loud[j-1])])
            startP = j
    chunks.append([startP, has_loud_length, int(has_loud[j])])
    return chunks


def apply_frame_margin(has_loud, has_loud_length, frame_margin):
    # (np.ndarray, int, int) -> np.ndarray:
    new = np.zeros((has_loud_length), dtype=np.uint8)
    for i in range(has_loud_length):
        start = int(max(0, i - frame_margin))
        end = int(min(has_loud_length, i+1+frame_margin))
        new[i] = min(1, np.max(has_loud[start:end]))
    return new

def apply_spacing_rules(has_loud, has_loud_length, minClip, minCut, speeds, fps, args, log):
    # (has_loud: np.ndarray, speeds: list, fps: float, args, log) -> list:
    if(args.mark_as_loud != []):
        has_loud = setRange(has_loud, args.mark_as_loud, fps, 1, log)

    if(args.mark_as_silent != []):
        has_loud = setRange(has_loud, args.mark_as_silent, fps, 0, log)

    # Remove small clips/cuts created by applying other rules.
    has_loud = cook(has_loud, minClip, minCut)

    if(args.cut_out != []):
        cut_speed_index = speeds.index(99999)
        has_loud = setRange(has_loud, args.cut_out, fps, cut_speed_index, log)

    if(args.set_speed_for_range != []):
        for item in args.set_speed_for_range:
            my_speed_index = speeds.index(float(item[0]))
            has_loud = setRange(has_loud, [item[1:]], fps, my_speed_index, log)

    return chunkify(has_loud, has_loud_length)


def apply_basic_spacing(has_loud, fps, minClip, minCut, log):
    # (has_loud: np.ndarray, fps: float, minClip: int, minCut: int, log) -> list:
    minClip = seconds_to_frames(minClip, fps)
    minCut = seconds_to_frames(minCut, fps)

    has_loud = cook(has_loud, minClip, minCut)
    return chunkify(has_loud)


def merge(start_list, end_list):
    # (start_list: np.ndarray, end_list: np.ndarray) -> np.ndarray:
    merge = np.zeros((len(start_list)), dtype=np.bool_)

    startP = 0
    for item in start_list:
        if(item == True):
            where_list = np.where(end_list[startP:])[0]
            if(len(where_list) > 0):
                merge[startP:where_list[0]] = True
        startP += 1
    return merge


def handleBoolExp(val, data, sampleRate, fps, log):
    # (val: str, data, sampleRate, fps, log) -> list:
    invert = False

    if('>' in val and '<' in val):
        log.error('Cannot have both ">" and "<" in same expression.')
    if('>' in val):
        exp = val.split('>')
    elif('<' in val):
        exp = val.split('<')
        invert = True
    else:
        log.error('audio array needs ">" or "<".')

    if(len(exp) != 2):
        log.error('Only one expression supported, not {}.'.format(len(exp)-1))

    if(data is None or sampleRate is None):
        log.error('No audio data found.')

    from auto_editor.analyze import audio_detection
    new_list = audio_detection(data, sampleRate, float(exp[1]), fps, log)

    if(invert):
        new_list = np.invert(new_list)

    return new_list


def applyRects(cmdRects, audioData, sampleRate, fps, log):
    from auto_editor.utils.func import hex_to_bgr

    rects = []
    for ms in cmdRects:
        if(len(ms) < 6):
            log.error('Too few comma arguments for rectangle option.')

        if(len(ms) > 8):
            log.error('Too many comma arguments for rectangle option.')

        start, end, x1, y1, x2, y2 = ms[:6]

        color = '#000'
        thickness = -1
        if(len(ms) > 6):
            color = ms[6]
        if(len(ms) > 7):
            thickness = int(ms[7])
        color = hex_to_bgr(color, log)

        # Handle Boolean Expressions. Mostly the same as zoom.
        start_list, end_list = None, None
        if(start.startswith('audio')):
            start_list = handleBoolExp(start, audioData, sampleRate, fps, log)

        if(end.startswith('audio')):
            if(start_list is None):
                log.error('The start parameter must also have a boolean expression.')
            end_list = handleBoolExp(end, audioData, sampleRate, fps, log)

        if(start_list is None):
            rects.append(['rectangle', start, end, x1, y1, x2, y2, color, thickness])

        elif(end_list is None):
            # Handle if end is not a boolean expression.
            indexs = np.where(start_list)[0]
            if(indexs != []):
                rects.append(['rectangle', str(indexs[0]), end, x1, y1, x2, y2, color,
                    thickness])
        else:
            chunks = apply_basic_spacing(merge(start_list, end_list), fps, 0, 0, log)
            for item in chunks:
                if(item[2] == 1):
                    rects.append(['rectangle', str(item[0]), str(item[1]), x1, y1, x2, y2,
                        color, thickness])

            if(rects == []):
                log.warning('No rectangles applied.')
            else:
                log.print(' {} rectangles applied.'.format(len(rects)))

    return rects


def applyZooms(cmdZooms, audioData, sampleRate, fps, log):
    zooms = []
    for ms in cmdZooms:

        start, end = ms[:2]

        start_zoom = float(ms[2])

        if(len(ms) == 3):
            end_zoom = start_zoom
        else:
            end_zoom = float(ms[3])

        x = 'centerX'
        y = 'centerY'
        inter = 'linear'
        hold = None

        if(len(ms) > 4):
            x, y = ms[4:6]

        if(len(ms) > 6):
            inter = ms[6]

        if(len(ms) > 7):
            hold = ms[7]

        start_list, end_list = None, None
        if(start.startswith('audio')):
            start_list = handleBoolExp(start, audioData, sampleRate, fps, log)

        if(end.startswith('audio')):
            if(start_list is None):
                log.error('The start parameter must also have a boolean expression.')
            end_list = handleBoolExp(end, audioData, sampleRate, fps, log)

        if(start_list is None):
            zooms.append(['zoom', start, end, start_zoom, end_zoom, x, y, inter, hold])

        elif(end_list is None):
            # Handle if end is not a boolean expression.
            indexs = np.where(start_list)[0]
            if(indexs != []):
                zooms.append(['zoom', str(indexs[0]), end, start_zoom, end_zoom, x, y,
                    inter, hold])
        else:
            chunks = apply_basic_spacing(merge(start_list, end_list), fps, 0, 0, log)
            for item in chunks:
                if(item[2] == 1):
                    zooms.append(['zoom', str(item[0]), str(item[1]), start_zoom,
                        end_zoom, x, y, inter, hold])

            if(zooms == []):
                log.warning('No zooms applied.')
            else:
                log.print(' {} zooms applied.'.format(len(zooms)))

    log.debug(zooms)
    return zooms



'''render/opencv.py'''

import cv2
import numpy as np

import os.path

# Included Libraries
from auto_editor.interpolate import interpolate
from auto_editor.utils.progressbar import ProgressBar
from .utils import properties, scale_to_sped

def values(val, log, _type, total_frames, width, height):
    if(val == 'centerX'):
        return int(width / 2)
    if(val == 'centerY'):
        return int(height / 2)
    if(val == 'start'):
        return 0
    if(val == 'end'):
        return total_frames - 1
    if(val == 'width'):
        return width
    if(val == 'height'):
        return height

    if(not isinstance(val, int)
        and not (val.replace('.', '', 1)).replace('-', '', 1).isdigit()):
        log.error('Variable {} not implemented.'.format(val))
    return _type(val)

def _find_state(chunks, cframe):
    low = 0
    high = len(chunks) - 1

    while low <= high:
        mid = low + (high - low) // 2

        if(cframe >= chunks[mid][0] and cframe < chunks[mid][1]):
            return chunks[mid][2]
        elif(cframe > chunks[mid][0]):
            low = mid + 1
        else:
            high = mid - 1

    # cframe not in chunks
    return 0


def make_effects_sheet(effects, total_frames, width, height, log):
    effect_sheet = []
    for effect in effects:
        if(effect[0] == 'rectangle'):

            rectx1_sheet = np.zeros((total_frames + 1), dtype=int)
            recty1_sheet = np.zeros((total_frames + 1), dtype=int)
            rectx2_sheet = np.zeros((total_frames + 1), dtype=int)
            recty2_sheet = np.zeros((total_frames + 1), dtype=int)
            rectco_sheet = np.zeros((total_frames + 1, 3), dtype=int)
            rect_t_sheet = np.zeros((total_frames + 1), dtype=int)

            r = effect[1:]

            for i in range(6):
                r[i] = values(r[i], log, int, total_frames, width, height)

            rectx1_sheet[r[0]:r[1]] = r[2]
            recty1_sheet[r[0]:r[1]] = r[3]
            rectx2_sheet[r[0]:r[1]] = r[4]
            recty2_sheet[r[0]:r[1]] = r[5]
            rectco_sheet[r[0]:r[1]] = r[6]
            rect_t_sheet[r[0]:r[1]] = r[7]

            effect_sheet.append(
                ['rectangle', rectx1_sheet, recty1_sheet, rectx2_sheet, recty2_sheet,
                rectco_sheet, rect_t_sheet]
            )

        if(effect[0] == 'zoom'):

            zoom_sheet = np.ones((total_frames + 1), dtype=float)
            zoomx_sheet = np.full((total_frames + 1), int(width / 2), dtype=float)
            zoomy_sheet = np.full((total_frames + 1), int(height / 2), dtype=float)

            z = effect[1:]
            z[0] = values(z[0], log, int, total_frames, width, height)
            z[1] = values(z[1], log, int, total_frames, width, height)

            if(z[7] is not None): # hold value
                z[7] = values(z[7], log, int, total_frames, width, height)

            if(z[7] is None or z[7] > z[1]):
                zoom_sheet[z[0]:z[1]] = interpolate(z[2], z[3], z[1] - z[0], log,
                    method=z[6])
            else:
                zoom_sheet[z[0]:z[0]+z[7]] = interpolate(z[2], z[3], z[7], log,
                    method=z[6])
                zoom_sheet[z[0]+z[7]:z[1]] = z[3]

            zoomx_sheet[z[0]:z[1]] = values(z[4], log, float, total_frames, width, height)
            zoomy_sheet[z[0]:z[1]] = values(z[5], log, float, total_frames, width, height)

            effect_sheet.append(
                ['zoom', zoom_sheet, zoomx_sheet, zoomy_sheet]
            )
    return effect_sheet

def render_opencv(ffmpeg, inp, args, chunks, speeds, fps, has_vfr, effects, temp, log):
    if(has_vfr):
        cmd = ['-i', inp.path, '-map', '0:v:0', '-vf', 'fps=fps={}'.format(fps), '-r',
            str(fps), '-vsync', '1', '-f','matroska', '-vcodec', 'rawvideo', 'pipe:1']
        fileno = ffmpeg.Popen(cmd).stdout.fileno()
        cap = cv2.VideoCapture('pipe:{}'.format(fileno))
    else:
        cap = cv2.VideoCapture(inp.path)

    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')

    if(args.scale != 1):
        width = int(width * args.scale)
        height = int(height * args.scale)
        video_name = os.path.join(temp, 'scale.mp4')
    else:
        video_name = os.path.join(temp, 'spedup.mp4')

    if(width < 2 or height < 2):
        log.error('Resolution too small.')
    log.debug('Resolution: {}x{}'.format(width, height))

    out = cv2.VideoWriter(video_name, fourcc, fps, (width, height))

    total_frames = chunks[len(chunks) - 1][1]
    cframe = 0

    cap.set(cv2.CAP_PROP_POS_FRAMES, cframe)
    remander = 0
    framesWritten = 0

    videoProgress = ProgressBar(total_frames, 'Creating new video',
        args.machine_readable_progress, args.no_progress)

    sheet = make_effects_sheet(effects, total_frames, width, height, log)

    while cap.isOpened():
        ret, frame = cap.read()
        if(not ret or cframe > total_frames):
            break

        for effect in sheet:
            if(effect[0] == 'rectangle'):

                x1 = int(effect[1][cframe])
                y1 = int(effect[2][cframe])
                x2 = int(effect[3][cframe])
                y2 = int(effect[4][cframe])

                if(x1 == y1 and y1 == x2 and x2 == y2 and y2 == 0):
                    pass
                else:
                    np_color = effect[5][cframe]
                    color = (int(np_color[0]), int(np_color[1]), int(np_color[2]))

                    t = int(effect[6][cframe])

                    frame = cv2.rectangle(frame, (x1,y1), (x2,y2), color, thickness=t)

            if(effect[0] == 'zoom'):

                zoom = effect[1][cframe]
                zoom_x = effect[2][cframe]
                zoom_y = effect[3][cframe]

                # Resize Frame
                new_size = (int(width * zoom), int(height * zoom))

                if(zoom == 1 and args.scale == 1):
                    blown = frame
                elif(new_size[0] < 1 or new_size[1] < 1):
                    blown = cv2.resize(frame, (1, 1), interpolation=cv2.INTER_AREA)
                else:
                    inter = cv2.INTER_CUBIC if zoom > 1 else cv2.INTER_AREA
                    blown = cv2.resize(frame, new_size, interpolation=inter)

                x1 = int((zoom_x * zoom)) - int((width / 2))
                x2 = int((zoom_x * zoom)) + int((width / 2))

                y1 = int((zoom_y * zoom)) - int((height / 2))
                y2 = int((zoom_y * zoom)) + int((height / 2))

                top, bottom, left, right = 0, 0, 0, 0

                if(y1 < 0):
                    top = -y1
                    y1 = 0
                if(x1 < 0):
                    left = -x1
                    x1 = 0

                frame = blown[y1:y2+1, x1:x2+1]

                bottom = (height + 1) - (frame.shape[0]) - top
                right = (width + 1) - frame.shape[1] - left
                frame = cv2.copyMakeBorder(
                    frame,
                    top = top,
                    bottom = bottom,
                    left = left,
                    right = right,
                    borderType = cv2.BORDER_CONSTANT,
                    value = args.background
                )

                if(frame.shape != (height+1, width+1, 3)):
                    # Throw error so that opencv dropped frames don't go unnoticed.
                    print('cframe {}'.format(cframe))
                    log.error('Wrong frame shape. Is {}, should be {}'.format(
                        frame.shape, (height+1, width+1, 3)))

        if(effects == [] and args.scale != 1):
            inter = cv2.INTER_CUBIC if args.scale > 1 else cv2.INTER_AREA
            frame = cv2.resize(frame, (width, height), interpolation=inter)

        cframe = int(cap.get(cv2.CAP_PROP_POS_FRAMES)) # current frame

        state = _find_state(chunks, cframe)
        mySpeed = speeds[state]

        if(mySpeed != 99999):
            doIt = (1 / mySpeed) + remander
            for __ in range(int(doIt)):
                out.write(frame)
                framesWritten += 1
            remander = doIt % 1

        videoProgress.tick(cframe)
    log.debug('Frames Written: {}'.format(framesWritten))
    log.debug('Total Frames: {}'.format(total_frames))

    cap.release()
    out.release()
    cv2.destroyAllWindows()

    if(args.scale == 1):
        cmd = properties(['-i', inp.path], args, inp)
        cmd.append(os.path.join(temp, 'spedup.mp4'))
        ffmpeg.run(cmd)
    else:
        scale_to_sped(ffmpeg, inp, args, temp)



'''render/audio.py'''

from __future__ import division

import os

from auto_editor.utils.func import get_new_length, fnone
from auto_editor.utils.progressbar import ProgressBar
from auto_editor.scipy.wavfile import read, write

def convertAudio(ffmpeg, in_file, inp, out_file, codec, log):
    if(fnone(codec)):
        codec = inp.audio_streams[0]['codec']
    if(codec == 'pcm_s16le' and out_file.endswith('.m4a')):
        log.error('Codec: {} is not supported in the m4a container.'.format(codec))

    ffmpeg.run(['-i', in_file, '-acodec', codec, out_file])

def handleAudio(ffmpeg, in_file, audioBit, samplerate, temp, log):
    # () -> str:
    temp_file = os.path.join(temp, 'faAudio.wav')

    log.checkType(samplerate, 'samplerate', str)
    cmd = ['-i', in_file]
    if(not fnone(audioBit)):
        cmd.extend(['-b:a', audioBit])
    cmd.extend(['-ac', '2', '-ar', samplerate, '-vn', temp_file])

    ffmpeg.run(cmd)
    log.conwrite('')

    return temp_file

def fastAudio(in_file, out_file, chunks, speeds, log, fps, machineReadable, hideBar):
    import numpy as np

    def custom_speeds(a):
        return len([x for x in a if x != 1 and x != 99999]) > 0

    if(custom_speeds(speeds)):
        from audiotsm2 import phasevocoder
        from audiotsm2.io.array import ArrReader, ArrWriter

    if(len(chunks) == 1 and chunks[0][2] == 0):
        log.error('Trying to create an empty file.')

    samplerate, audioData = read(in_file)

    newL = get_new_length(chunks, speeds, fps)
    # Get the new length in samples with some extra leeway.
    estLeng = int(newL * samplerate * 1.5) + int(samplerate * 2)

    # Create an empty array for the new audio.
    newAudio = np.zeros((estLeng, 2), dtype=np.int16)

    channels = 2
    yPointer = 0

    audioProgress = ProgressBar(len(chunks), 'Creating new audio', machineReadable,
        hideBar)

    for chunkNum, chunk in enumerate(chunks):
        audioSampleStart = int(chunk[0] / fps * samplerate)
        audioSampleEnd = int(audioSampleStart + (samplerate / fps) * (chunk[1] - chunk[0]))

        theSpeed = speeds[chunk[2]]
        if(theSpeed != 99999):
            spedChunk = audioData[audioSampleStart:audioSampleEnd]

            if(theSpeed == 1):
                yPointerEnd = yPointer + spedChunk.shape[0]
                newAudio[yPointer:yPointerEnd] = spedChunk
            else:
                spedupAudio = np.zeros((0, 2), dtype=np.int16)
                with ArrReader(spedChunk, channels, samplerate, 2) as reader:
                    with ArrWriter(spedupAudio, channels, samplerate, 2) as writer:
                        phasevocoder(reader.channels, speed=theSpeed).run(
                            reader, writer
                        )
                        spedupAudio = writer.output

                yPointerEnd = yPointer + spedupAudio.shape[0]
                newAudio[yPointer:yPointerEnd] = spedupAudio

            myL = chunk[1] - chunk[0]
            mySamples = (myL / fps) * samplerate
            newSamples = int(mySamples / theSpeed)

            yPointer = yPointer + newSamples
        else:
            # Speed is too high so skip this section.
            yPointerEnd = yPointer

        audioProgress.tick(chunkNum)

    log.debug('Total Samples: {}'.format(yPointer))
    log.debug('Samples per Frame: {}'.format(samplerate / fps))
    log.debug('Expected video length: {}'.format(yPointer / (samplerate / fps)))
    newAudio = newAudio[:yPointer]
    write(out_file, samplerate, newAudio)

    log.conwrite('')



'''utils/video.py'''

# Internal Libraries
import os
from shutil import move, copy

# Included Libraries
from .func import fnone
from auto_editor.render.audio import fastAudio

def handleAudioTracks(ffmpeg, write_file, args, tracks, chunks, speeds, fps, temp, log):
    for t in range(tracks):
        temp_file = os.path.join(temp, '{}.wav'.format(t))
        new_file = os.path.join(temp, 'new{}.wav'.format(t))
        fastAudio(temp_file, new_file, chunks, speeds, log, fps,
            args.machine_readable_progress, args.no_progress)

        if(not os.path.isfile(new_file)):
            log.bug('Audio file not created.')

    if(args.export_as_audio):
        if(args.keep_tracks_seperate):
            log.warning("Audio files don't have multiple tracks.")

        if(tracks == 1):
            move(os.path.join(temp, '0.wav'), write_file)
            return False

        cmd = []
        for t in range(tracks):
            cmd.extend(['-i', os.path.join(temp, '{}.wav'.format(t))])
        cmd.extend(['-filter_complex', 'amix=inputs={}:duration=longest'.format(tracks),
            write_file])
        ffmpeg.run(cmd)

        return False
    return True

def mux_rename_video(ffmpeg, write_file, args, tracks, temp, log):

    if(tracks == 0):
        move(os.path.join(temp, 'spedup.mp4'), write_file)
        log.conwrite('')
        return

    cmd = []
    if(args.keep_tracks_seperate):
        for t in range(tracks):
            cmd.extend(['-i', os.path.join(temp, 'new{}.wav'.format(t))])
        cmd.extend(['-i', os.path.join(temp, 'spedup.mp4')])
        for t in range(tracks):
            cmd.extend(['-map', '{}:a:0'.format(t)])
        cmd.extend(['-map', '{}:v:0'.format(tracks)])
    else:
        # Merge all the audio tracks into one.
        new_a_file = os.path.join(temp, 'newAudioFile.wav')
        if(tracks > 1):
            for t in range(tracks):
                cmd.extend(['-i', os.path.join(temp, 'new{}.wav'.format(t))])
            cmd.extend(['-filter_complex', 'amerge=inputs={}'.format(tracks), '-ac', '2',
                new_a_file])
            ffmpeg.run(cmd)
        else:
            copy(os.path.join(temp, 'new0.wav'), new_a_file)
        cmd = ['-i', new_a_file, '-i', os.path.join(temp, 'spedup.mp4')]

    cmd.extend(['-c:v', 'copy'])
    if(not fnone(args.audio_codec)):
        cmd.extend(['-c:a', args.audio_codec])
    cmd.append(write_file)
    ffmpeg.run(cmd)
    log.conwrite('')



'''subcommands/grep.py'''

from __future__ import print_function

def grep_options(parser):
    parser.add_argument('--max_count', '-m', type=int, default=float('inf'),
        help='Stop reading a file after NUM matching lines')
    parser.add_argument('--count', '-c', action='store_true',
        help='Suppress normal output; instead print count of matching lines for each file.')
    parser.add_argument('--ignore_case', '-i', action='store_true',
        help='Ignore case distinctions for the PATTERN.')
    parser.add_argument('--ffmpeg_location', default=None,
        help='point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    parser.add_argument('input', nargs='*',
        help='the path to a file you want inspected.')
    return parser


def grep(sys_args=None):
    import os
    import re
    import sys
    import tempfile

    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.log import Log
    from auto_editor.ffwrapper import FFmpeg

    parser = vanparse.ArgumentParser('grep', auto_editor.version,
        description='Read and match subtitle tracks in media files.')
    parser = grep_options(parser)

    if(sys_args is None):
        sys_args = sys.args[1:]

    TEMP = tempfile.mkdtemp()
    log = Log(temp=TEMP)
    args = parser.parse_args(sys_args, log, 'grep')

    dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    ffmpeg = FFmpeg(dir_path, args.ffmpeg_location, args.my_ffmpeg, False, log)

    regex = args.input[0]
    media_files = args.input[1:]

    flags = 0
    if(args.ignore_case):
        flags = re.IGNORECASE

    """
    we're using the WEBVTT subtitle format. It's better than srt
    because it doesn't emit line numbers and the time code is in
    (hh:mm:ss.sss) instead of (dd:hh:mm:ss,sss)
    """

    for media_file in media_files:

        if(not os.path.exists(media_file)):
            log.error('{}: File does not exist.'.format(media_file))

        out_file = os.path.join(TEMP, 'media.vtt')
        ffmpeg.run(['-i', media_file, out_file])

        count = 0

        prefix = ''
        if(len(media_files) > 1):
            prefix = '{}:'.format(media_file)

        with open(out_file, 'r') as file:
            while True:
                line = file.readline()

                if(not line or count >= args.max_count):
                    break

                match = re.search(regex, line, flags)

                if(match):
                    count += 1
                    if(not args.count):
                        print(prefix + line.strip())

        if(args.count):
            print(prefix + str(count))

    log.cleanup()

if(__name__ == '__main__'):
    grep()



'''formats/final_cut_pro.py'''

from .utils import indent, get_width_height

def fcp_xml(inp, temp, output, clips, tracks, total_dur, audioFile, fps, log):

    pathurl = 'file://' + inp.abspath
    name = inp.basename

    def fraction(a, fps) -> str:
        from fractions import Fraction

        if(a == 0):
            return '0s'

        if(isinstance(a, float)):
            a = Fraction(a)
        if(isinstance(fps, float)):
            fps = Fraction(fps)

        frac = Fraction(a, fps).limit_denominator()
        num = frac.numerator
        dem = frac.denominator

        if(dem < 3000):
            factor = int(3000 / dem)

            if(factor == 3000 / dem):
                num *= factor
                dem *= factor
            else:
                # Good enough but has some error that are impacted at speeds such as 150%.
                total = 0
                while(total < frac):
                    total += Fraction(1, 30)
                num = total.numerator
                dem = total.denominator

        return '{}/{}s'.format(num, dem)

    width, height = get_width_height(inp)
    if(width is None or height is None):
        width, height = '1280', '720'
    frame_duration = fraction(1, fps)

    with open(output, 'w', encoding='utf-8') as outfile:
        outfile.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        outfile.write('<!DOCTYPE fcpxml>\n\n')
        outfile.write('<fcpxml version="1.9">\n')
        outfile.write('\t<resources>\n')
        outfile.write('\t\t<format id="r1" name="FFVideoFormat{}p{}" '.format(height, fps)+\
            'frameDuration="{}" width="{}" height="{}"'.format(frame_duration, width, height)+\
            ' colorSpace="1-1-1 (Rec. 709)"/>\n')

        outfile.write('\t\t<asset id="r2" name="{}" start="0s" '.format(name) +
            'hasVideo="1" format="r1" hasAudio="1" ' +
            'audioSources="1" audioChannels="2" duration="{}">\n'.format(
                fraction(total_dur, fps)))

        outfile.write('\t\t\t<media-rep kind="original-media" ' +
            'src="{}"></media-rep>\n'.format(pathurl))
        outfile.write('\t\t</asset>\n')
        outfile.write('\t</resources>\n')
        outfile.write('\t<library>\n')
        outfile.write('\t\t<event name="auto-editor output">\n')
        outfile.write('\t\t\t<project name="{}">\n'.format(name))
        outfile.write(indent(4,
            '<sequence format="r1" tcStart="0s" tcFormat="NDF" '\
            'audioLayout="stereo" audioRate="48k">',
            '\t<spine>')
        )

        last_dur = 0
        for clip in clips:
            clip_dur = (clip[1] - clip[0]) / (clip[2] / 100)
            dur = fraction(clip_dur, fps)

            close = '/' if clip[2] == 100 else ''

            if(last_dur == 0):
                outfile.write(indent(6, '<asset-clip name="{}" offset="0s" ref="r2"'.format(name)+\
                ' duration="{}" tcFormat="NDF"{}>'.format(dur, close)))
            else:
                start = fraction(clip[0] / (clip[2] / 100), fps)
                off = fraction(last_dur, fps)
                outfile.write(indent(6,
                    '<asset-clip name="{}" offset="{}" ref="r2" '.format(name, off)+\
                    'duration="{}" start="{}" '.format(dur, start)+\
                    'tcFormat="NDF"{}>'.format(close),
                ))

            if(clip[2] != 100):
                # See the "Time Maps" section.
                # https://developer.apple.com/library/archive/documentation/FinalCutProX
                #    /Reference/FinalCutProXXMLFormat/StoryElements/StoryElements.html

                frac_total = fraction(total_dur, fps)
                total_dur_divided_by_speed = fraction((total_dur) / (clip[2] / 100), fps)

                outfile.write(indent(6,
                    '\t<timeMap>',
                    '\t\t<timept time="0s" value="0s" interp="smooth2"/>',
                    '\t\t<timept time="{}" value="{}" interp="smooth2"/>'.format(total_dur_divided_by_speed, frac_total),
                    '\t</timeMap>',
                    '</asset-clip>'
                ))

            last_dur += clip_dur

        outfile.write('\t\t\t\t\t</spine>\n')
        outfile.write('\t\t\t\t</sequence>\n')
        outfile.write('\t\t\t</project>\n')
        outfile.write('\t\t</event>\n')
        outfile.write('\t</library>\n')
        outfile.write('</fcpxml>\n')



'''validateInput.py'''

from __future__ import print_function

import os
import re
import sys

from auto_editor.utils.progressbar import ProgressBar

invalidExtensions = ['.txt', '.md', '.rtf', '.csv', '.cvs', '.html', '.htm',
      '.xml', '.yaml', '.png', '.jpeg', '.jpg', '.exe', '.doc',
      '.docx', '.odt', '.pptx', '.xlsx', '.xls', 'ods', '.pdf', '.bat', '.dll',
      '.prproj', '.psd', '.aep', '.zip', '.rar', '.7z', '.java', '.class', '.js',
      '.c', '.cpp', '.csharp', '.py', '.app', '.git', '.github', '.gitignore',
      '.db', '.ini', '.BIN', '.svg', '.in', '.pyc', '.log', '.xsd', '.ffpreset',
      '.kys', '.essentialsound']

class MyLogger(object):
    @staticmethod
    def debug(msg):
        pass

    @staticmethod
    def warning(msg):
        print(msg, file=sys.stderr)

    @staticmethod
    def error(msg):
        if("'Connection refused'" in msg):
            pass
        else:
            print(msg, file=sys.stderr)


def parse_bytes(bytestr):
    # Parse a string indicating a byte quantity into an integer.
    matchobj = re.match(r'(?i)^(\d+(?:\.\d+)?)([kMGTPEZY]?)$', bytestr)
    if(matchobj is None):
        return None
    number = float(matchobj.group(1))
    multiplier = 1024.0 ** 'bkmgtpezy'.index(matchobj.group(2).lower())
    return int(round(number * multiplier))


def sponsor_block_api(_id, categories, log):
    # (_id: str, categories: list, log) -> dict
    from urllib import request
    from urllib.error import HTTPError
    import json

    cat_url = 'categories=['
    for i, cat in enumerate(categories):
        if(i == 0):
            cat_url += '"{}"'.format(cat)
        else:
            cat_url += ',"{}"'.format(cat)
    cat_url += ']'

    try:
        contents = request.urlopen(
            'https://sponsor.ajay.app/api/skipSegments?videoID={}&{}'.format(_id, cat_url))
        return json.loads(contents.read())
    except HTTPError:
        log.warning("Couldn't find skipSegments for id: {}".format(_id))
        return None

def download_video(my_input, args, ffmpeg, log):
    outtmpl = re.sub(r'\W+', '-', my_input)
    if(outtmpl.endswith('-mp4')):
        outtmpl = outtmpl[:-4]
    outtmpl += '.mp4'

    if(args.output_dir is not None):
        outtmpl = os.path.join(args.output_dir, outtmpl)

    try:
        import youtube_dl
    except ImportError:
        log.error('Download the youtube-dl python library to download URLs.\n' \
            '   pip3 install youtube-dl')

    if(not os.path.isfile(outtmpl)):
        ytbar = ProgressBar(100, 'Downloading')
        def my_hook(d):
            if(d['status'] == 'downloading'):
                ytbar.tick(float(d['_percent_str'].replace('%','')))

        ydl_opts = {
            'nocheckcertificate': not args.check_certificate,
            'outtmpl': outtmpl,
            'ffmpeg_location': ffmpeg.getPath(),
            'format': args.format,
            'ratelimit': parse_bytes(args.limit_rate),
            'logger': MyLogger(),
            'cookiefile': os.path.abspath(args.cookies),
            'download_archive': os.path.abspath(args.download_archive),
            'progress_hooks': [my_hook],
        }

        with youtube_dl.YoutubeDL(ydl_opts) as ydl:
            try:
                ydl.download([my_input])
            except youtube_dl.utils.DownloadError:
                log.conwrite('')
                log.error('YouTube-dl: Connection Refused.')

        log.conwrite('')
    return outtmpl

def _valid_files(path, bad_exts):
    # (path: str, badExts: list)
    for f in os.listdir(path):
        if(f[f.rfind('.'):] not in bad_exts and not os.path.isdir(f)
            and not f.startswith('.')):
            yield os.path.join(path, f)

def get_segment(args, my_input, log):
    if(args.block is not None):
        if(args.id is not None):
            return sponsor_block_api(args.id, args.block, log)
        match = re.search(r'youtube\.com/watch\?v=(?P<match>[A-Za-z0-9_-]{11})',
            my_input)
        if(match):
            youtube_id =  match.groupdict()['match']
            return sponsor_block_api(youtube_id, args.block, log)
    return None

def valid_input(inputs, ffmpeg, args, log):
    # (inputs: list, ffmpeg, args, log) -> list:
    new_inputs = []
    segments = []
    for my_input in inputs:
        if(os.path.isdir(my_input)):
            new_inputs += sorted(_valid_files(my_input, invalidExtensions))
            segments += [None] * (len(new_inputs) - len(segments))
        elif(os.path.isfile(my_input)):
            _, ext = os.path.splitext(my_input)
            if(ext == ''):
                log.error('File must have an extension.')

            if(ext in invalidExtensions):
                log.error('Invalid file extension "{}" for {}'.format(ext, my_input))
            new_inputs.append(my_input)
            segments.append(get_segment(args, my_input, log))

        elif(my_input.startswith('http://') or my_input.startswith('https://')):
            new_inputs.append(download_video(my_input, args, ffmpeg, log))
            segments.append(get_segment(args, my_input, log))
        else:
            log.error('Could not find file: {}'.format(my_input))

    return new_inputs, segments



'''render/utils.py'''

import os.path

from auto_editor.utils.func import fnone

def fset(cmd, option, value):
    if(fnone(value)):
        return cmd
    return cmd + [option] + [value]

def properties(cmd, args, inp):
    if(args.video_codec == 'uncompressed'):
        cmd.extend(['-vcodec', 'mpeg4', '-qscale:v', '1'])
    elif(inp.ext == '.gif'):
        cmd.extend(['-vcodec', 'gif'])
    elif(args.video_codec == 'copy'):
        new_codec = inp.video_streams[0]['codec']
        if(new_codec != 'dvvideo'): # This codec seems strange.
            cmd.extend(['-vcodec', new_codec])
    else:
        cmd = fset(cmd, '-vcodec', args.video_codec)

    cmd = fset(cmd, '-crf', args.constant_rate_factor)
    cmd = fset(cmd, '-b:v', args.video_bitrate)
    cmd = fset(cmd, '-tune', args.tune)
    cmd = fset(cmd, '-preset', args.preset)

    cmd.extend(['-movflags', '+faststart', '-strict', '-2'])
    return cmd

def scale_to_sped(ffmpeg, inp, args, temp):
    SCALE = os.path.join(temp, 'scale.mp4')
    SPED = os.path.join(temp, 'spedup.mp4')

    cmd = ['-i', SCALE]
    cmd = properties(cmd, args, inp)
    cmd.append(SPED)
    check_errors = ffmpeg.pipe(cmd)

    if('Error' in check_errors or 'failed' in check_errors):
        cmd = ['-i', SCALE]
        if('-allow_sw 1' in check_errors):
            cmd.extend(['-allow_sw', '1'])

        cmd = properties(cmd, args, inp)
        cmd.append(SPED)
        ffmpeg.run(cmd)



'''analyze.py'''

from __future__ import division

import math

import numpy as np

def get_blank_list(inp, audioData, sampleRate, fps):
    if(audioData is not None):
        audioSampleCount = audioData.shape[0]
        samplesPerFrame = sampleRate / fps
        audioFrameCount = int(math.ceil(audioSampleCount / samplesPerFrame))
        return np.ones((audioFrameCount), dtype=np.bool_)

    import cv2
    cap = cv2.VideoCapture(inp.path)
    totalFrames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) + 1
    return np.ones((totalFrames), dtype=np.bool_)

def audio_detection(audioData, sampleRate, silent_threshold, fps, log):
    # (audioData: np.ndarray, sampleRate: int, silent_threshold: float,
    #    fps: float, log) -> np.ndarray:
    log.debug('Analyzing audio volume.')

    def getMaxVolume(s):
        # (s: np.ndarray) -> float:
        maxv = float(np.max(s))
        minv = float(np.min(s))
        return max(maxv, -minv)

    maxAudioVolume = getMaxVolume(audioData)
    audioSampleCount = audioData.shape[0]

    samplesPerFrame = sampleRate / fps
    audioFrameCount = int(math.ceil(audioSampleCount / samplesPerFrame))
    hasLoudAudio = np.zeros((audioFrameCount), dtype=np.bool_)

    if(maxAudioVolume == 0):
        log.error('The entire audio is completely silent.')

    # Calculate when the audio is loud or silent.
    for i in range(audioFrameCount):
        start = int(i * samplesPerFrame)
        end = min(int((i+1) * samplesPerFrame), audioSampleCount)
        audiochunks = audioData[start:end]
        if(getMaxVolume(audiochunks) / maxAudioVolume >= silent_threshold):
            hasLoudAudio[i] = True

    return hasLoudAudio


def motion_detection(inp, motionThreshold, log, width, dilates, blur):
    # (inp, motionThreshold: float, log, width: int, dilates: int, blur: int) -> np.ndarray:

    # Based on this post:
    # https://pyimagesearch.com/2015/05/25/basic-motion-detection-and-tracking-with-python-and-opencv/

    import cv2
    from auto_editor.utils.progressbar import ProgressBar

    log.debug('Analyzing video motion.')

    cap = cv2.VideoCapture(inp.path)

    totalFrames = int(cap.get(cv2.CAP_PROP_FRAME_COUNT)) + 1

    log.debug('   - Cutting totalFrames: {}'.format(totalFrames))
    prevFrame = None
    gray = None
    hasMotion = np.zeros((totalFrames), dtype=np.bool_)
    total = None

    def resize(image, width=None, height=None, inter=cv2.INTER_AREA):
        if(width is None and height is None):
            return image

        h, w = image.shape[:2]
        if(width is None):
            r = height / h
            dim = (int(w * r), height)
        else:
            r = width / w
            dim = (width, int(h * r))

        return cv2.resize(image, dim, interpolation=inter)

    progress = ProgressBar(totalFrames, 'Detecting motion')

    while cap.isOpened():
        if(gray is None):
            prevFrame = None
        else:
            prevFrame = gray

        ret, frame = cap.read()

        if(not ret):
            break

        cframe = int(cap.get(cv2.CAP_PROP_POS_FRAMES)) # current frame

        frame = resize(frame, width=width)
        gray = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY) # Convert frame to grayscale.
        if(blur > 0):
            gray = cv2.GaussianBlur(gray, (blur, blur), 0)

        if(prevFrame is not None):
            frameDelta = cv2.absdiff(prevFrame, gray)
            thresh = cv2.threshold(frameDelta, 25, 255, cv2.THRESH_BINARY)[1]

            # Dilate the thresholded image to fill in holes.
            if(dilates > 0):
                thresh = cv2.dilate(thresh, None, iterations=dilates)

            if(total is None):
                total = thresh.shape[0] * thresh.shape[1]

            if(np.count_nonzero(thresh) / total >= motionThreshold):
                hasMotion[cframe] = True

        progress.tick(cframe)

    cap.release()
    cv2.destroyAllWindows()

    log.conwrite('')

    return hasMotion



'''setup.py'''

import os
import re
import sys
from setuptools import setup, find_packages

def pip_version():
    with open(os.path.abspath('auto_editor/__init__.py')) as f:
        version_content = f.read()

    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_content, re.M)

    if(version_match):
        print(version_match)
        return version_match.group(1)

    raise ValueError('Unable to find version string.')


if(sys.argv[-1] == 'publish'):

    from shutil import rmtree
    rmtree('build')
    rmtree('dist')

    os.system('python3 setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()

with open('README.md', 'r') as f:
    long_description = f.read()

setup(
    name='auto-editor',
    version=pip_version(),
    description='Auto-Editor: Effort free video editing!',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='Unlicense',
    url='https://github.com/WyattBlue/auto-editor',
    author='WyattBlue',
    author_email='WyattBlue@protonmail.com',
    keywords='video audio media editor editing processing nonlinear automatic ' \
     'silence-detect silence-removal silence-speedup motion-detection',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'numpy',
        'audiotsm2',
        'opencv-python>=4.3',
        'youtube-dl',
        'av',
    ],
    classifiers=[
        'Topic :: Multimedia :: Video',
        'License :: Public Domain',
        'License :: OSI Approved :: The Unlicense (Unlicense)',
        'Environment :: Console',
        'Natural Language :: English',
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python :: Implementation :: IronPython',
        'Programming Language :: Python :: Implementation :: Jython',
    ],
    entry_points={
        "console_scripts": ["auto-editor=auto_editor.__main__:main"]
    }
)



'''vanparse.py'''

from __future__ import print_function

import os
import sys
import difflib

def out(text):
    import re
    import textwrap

    try:
        from shutil import get_terminal_size
        width = get_terminal_size().columns - 3
    except ImportError:
        width = 77

    indent_regex = re.compile(r'^(\s+)')
    wrapped_lines = []

    for line in text.split('\n'):
        exist_indent = re.search(indent_regex, line)
        pre_indent = exist_indent.groups()[0] if exist_indent else ''

        wrapped_lines.append(
            textwrap.fill(line, width=width, subsequent_indent=pre_indent)
        )

    print('\n'.join(wrapped_lines))

def print_option_help(args, option):
    text = ''
    if(option['action'] == 'grouping'):
        text += '  {}:\n'.format(option['names'][0])
    else:
        text += '  ' + ', '.join(option['names']) + '\n    ' + option['help'] + '\n\n'
        if(option['extra'] != ''):
            text += '\n{}\n\n'.format(option['extra'])

    if(option['action'] == 'default'):
        text += '    type: ' + option['type'].__name__
        text += '\n    default: {}\n'.format(option['default'])
        if(option['range'] is not None):
            text += '    range: ' +  option['range'] + '\n'

        if(option['choices'] is not None):
            text += '    choices: ' +  ', '.join(option['choices']) + '\n'
    elif(option['action'] == 'grouping'):
        for options in args:
            for op in options:
                if(op['group'] == option['names'][0]):
                    text += '  ' + ', '.join(op['names']) + ': ' + op['help'] + '\n'
    elif(option['action'] in ['store_true', 'store_false']):
        text += '    type: flag\n'
    else:
        text += '    type: unknown\n'

    if(option['group'] is not None):
        text += '    group: ' + option['group'] + '\n'
    out(text)

def print_program_help(root, the_args):
    text = ''
    for options in the_args:
        for option in options:
            if(option['action'] == 'grouping'):
                text += "\n  {}:\n".format(option['names'][0])
            else:
                text += '  ' + ', '.join(option['names']) + ': ' + option['help'] + '\n'
    text += '\n'
    if(root == 'auto-editor'):
        text += '  Have an issue? Make an issue. Visit '\
            'https://github.com/wyattblue/auto-editor/issues\n\n  The help option '\
            'can also be used on a specific option:\n     auto-editor '\
            '--frame_margin --help\n'
    out(text)

def get_option(item, group, the_args):
    for options in the_args:
        for option in options:
            if(item in option['names'] and group in ['global', option['group']]):
                return option
    return None

def _to_key(val):
    # (val: dict) -> str
    return val['names'][0].replace('-', '')

class ArgumentParser():
    def __init__(self, program_name, version, description=None):
        self.program_name = program_name
        self._version = version
        self.description = description

        self.args = []
        self.kwarg_defaults = {
            'nargs': 1,
            'type': str,
            'default': None,
            'action': 'default',
            'range': None,
            'choices': None,
            'group': None,
            'help': '',
            'extra': '',
        }

    def add_argument(self, *args, **kwargs):
        my_dict = {
            'names': list(args),
        }

        for key, item in self.kwarg_defaults.items():
            my_dict[key] = item

        for key, item in kwargs.items():
            if(key not in self.kwarg_defaults):
                raise ValueError('key {} not found.'.format(key))
            my_dict[key] = item

        self.args.append(my_dict)

    def parse_args(self, sys_args, log, root):
        if(sys_args == [] and self.description):
            out(self.description)
            sys.exit()

        if(sys_args == ['-v'] or sys_args == ['-V']):
            out('{} version {}'.format(self.program_name, self._version))
            sys.exit()

        return ParseOptions(sys_args, log, root, self.args)


class ParseOptions():

    def setConfig(self, config_path, root):
        if(not os.path.isfile(config_path)):
            return

        with open(config_path, 'r') as file:
            lines = file.readlines()

        # Set attributes based on the config file to act as the new defaults.
        for item in lines:
            if('#' in item):
                item = item[: item.index('#')]
            item = item.replace(' ', '')
            if(item.strip() == '' or (not item.startswith(root))):
                continue
            value = item[item.index('=')+1 :]

            if(value[0] == "'" and value[-1] == "'"):
                value = value[1:-1]
            elif(value == 'None'):
                value = None
            elif('.' in value):
                value = float(value)
            else:
                value = int(value)

            key = item[: item.index('=')]
            key = key[key.rfind('.')+1:]

            if(getattr(self, key) != value):
                print('Setting {} to {}'.format(key, value), file=sys.stderr)
            setattr(self, key, value)

    def __init__(self, sys_args, log, root, *args):
        # Set the default options.
        option_names = []
        for options in args:
            for option in options:
                option_names.append(option['names'][0])
                key = _to_key(option)
                if(option['action'] == 'store_true'):
                    value = False
                elif(option['action'] == 'store_false'):
                    value = True
                elif(option['nargs'] != 1):
                    value = []
                else:
                    value = option['default']
                setattr(self, key, value)

        dirPath = os.path.dirname(os.path.realpath(__file__))
        self.setConfig(os.path.join(dirPath, 'config.txt'), root)

        # Figure out command line options changed by user.
        my_list = []
        used_options = []
        _set = []
        setting_inputs = True
        option_list = 'input'
        list_type = str
        i = 0
        group = None
        while i < len(sys_args):
            item = sys_args[i]
            label = 'option' if item.startswith('--') else 'short'

            # Find the option.
            if(label == 'option'):
                option = get_option(item, group='global', the_args=args)
            else:
                option = get_option(item, group=group, the_args=args)
                if(option is None and (group is not None)):
                    group = None # Don't consider the next option to be in a group.
                    option = get_option(item, group=group, the_args=args)

            def error_message(args, item, label):

                def all_names(args):
                    name_set = set()
                    for options in args:
                        for opt in options:
                            for names in opt['names']:
                                name_set.add(names)
                    return name_set

                opt_list = all_names(args)
                close_matches = difflib.get_close_matches(item, opt_list)
                if(close_matches):
                    if(close_matches[0] == item):
                        return '{} {} needs to be in a group'.format(label.capitalize(),
                            item)
                    return 'Unknown {}: {}\n\n    Did you mean:\n        '.format(
                        label, item) + ', '.join(close_matches)
                return 'Unknown {}: {}'.format(label, item)

            if(option is None):
                # Unknown Option!
                if(setting_inputs and not item.startswith('-')):
                    # Option is actually an input file, like example.mp4
                    my_list.append(item)
                else:
                    log.error(error_message(args, item, label))
            else:
                # We found the option.
                if(option_list is not None):
                    setattr(self, option_list, list(map(list_type, my_list)))

                setting_inputs = False
                option_list = None
                my_list = []

                if(option['names'][0] in used_options):
                    log.error('Cannot repeat option {} twice.'.format(option['names'][0]))

                used_options.append(option['names'][0])

                key = _to_key(option)
                _set.append(key)

                if(option['action'] == 'grouping'):
                    group = key

                nextItem = None if i == len(sys_args) - 1 else sys_args[i+1]
                if(nextItem == '-h' or nextItem == '--help'):
                    print_option_help(args, option)
                    sys.exit()

                if(option['nargs'] != 1):
                    setting_inputs = True
                    option_list = key
                    list_type = option['type']
                elif(option['action'] == 'store_true'):
                    value = True
                elif(option['action'] == 'store_false'):
                    value = False
                else:
                    try:
                        value = option['type'](nextItem)
                    except Exception:
                        typeName = option['type'].__name__
                        log.error('Couldn\'t convert {} to {}'.format(
                            nextItem, typeName))

                    # Handle when the option value is not in choices list.
                    if(option['choices'] is not None and value not in option['choices']):
                        option_name = option['names'][0]
                        my_choices = ', '.join(option['choices'])
                        log.error('{} is not a choice for {}\nchoices are:\n  {}'.format(
                            value, option_name, my_choices))
                    i += 1
                setattr(self, key, value)

            i += 1
        if(setting_inputs):
            setattr(self, option_list, list(map(list_type, my_list)))
        setattr(self, '_set', _set)
        if(self.help):
            print_program_help(root, args)
            sys.exit()



'''subcommands/levels.py'''

from __future__ import print_function

def levels_options(parser):
    parser.add_argument('--output_file', '--output', '-o', type=str,
        default='data.txt')
    parser.add_argument('--track', type=int, default=0,
        help='what audio/video track to get. If --kind is set to motion, track will look at video streams.')
    parser.add_argument('--ffmpeg_location', default=None,
        help='point to your custom ffmpeg file.')
    parser.add_argument('--my_ffmpeg', action='store_true',
        help='use the ffmpeg on your PATH instead of the one packaged.')
    parser.add_argument('--help', '-h', action='store_true',
        help='print info about the program or an option and exit.')
    parser.add_argument('(input)', nargs='*',
        help='the template')
    return parser

def levels(sys_args=None):
    import os
    import sys
    import math
    import tempfile

    import numpy as np

    import auto_editor
    import auto_editor.vanparse as vanparse

    from auto_editor.utils.log import Log
    from auto_editor.ffwrapper import FFmpeg
    from auto_editor.scipy.wavfile import read

    parser = vanparse.ArgumentParser('levels', auto_editor.version,
        description='Get loudness of audio over time.')
    parser = levels_options(parser)

    if(sys_args is None):
        sys_args = sys.args[1:]

    temp = tempfile.mkdtemp()
    log = Log(temp=temp)

    args = parser.parse_args(sys_args, log, 'levels')

    dir_path = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    ffmpeg = FFmpeg(dir_path, args.ffmpeg_location, args.my_ffmpeg, False, log)

    inp = ffmpeg.file_info(args.input[0])

    fps = 30 if inp.fps is None else float(inp.fps)

    if(args.track > len(inp.audio_streams)):
        log.error('Audio track {} does not exist.'.format(args.track))

    # Split audio tracks into: 0.wav, 1.wav, etc.
    for t in range(len(inp.audio_streams)):
        ffmpeg.run(['-i', inp.path, '-ac', '2', '-map', '0:a:{}'.format(t),
            os.path.join(temp, '{}.wav'.format(t))])

    sample_rate, audio_data = read(os.path.join(temp, '{}.wav'.format(args.track)))
    audio_sample_count = audio_data.shape[0]

    def get_max_volume(s):
        return max(float(np.max(s)), -float(np.min(s)))

    max_volume = get_max_volume(audio_data)

    samples_per_frame = sample_rate / fps
    audio_frame_count = int(math.ceil(audio_sample_count / samples_per_frame))

    with open(args.output_file, 'w') as out:
        for i in range(audio_frame_count):
            start = int(i * samples_per_frame)
            end = min(int((i+1) * samples_per_frame), audio_sample_count)
            audiochunks = audio_data[start:end]
            out.write('{}\n'.format(get_max_volume(audiochunks) / max_volume))

    log.cleanup()

if(__name__ == '__main__'):
    levels()



'''utils/log.py'''

from __future__ import print_function

import sys
from shutil import rmtree
from time import time, sleep

from .func import term_size

class Timer():
    def __init__(self, quiet=False):
        self.start_time = time()
        self.quiet = quiet

    def stop(self):
        from datetime import timedelta

        second_len = round(time() - self.start_time, 2)
        minute_len = timedelta(seconds=round(second_len))
        if(not self.quiet):
            print('Finished. took {} seconds ({})'.format(second_len, minute_len))

class Log():
    def __init__(self, show_debug=False, quiet=False, temp=None):
        self.is_debug = show_debug
        self.quiet = quiet
        self.temp = temp

    def debug(self, message):
        if(self.is_debug):
            self.conwrite('')
            print('Debug: {}'.format(message))

    def cleanup(self):
        if(self.temp is None):
            return
        try:
            rmtree(self.temp)
            self.debug('Removed Temp Directory.')
        except PermissionError:
            sleep(0.1)
            try:
                rmtree(self.temp)
                self.debug('Removed Temp Directory.')
            except Exception:
                self.debug('Failed to delete temp dir.')

    def conwrite(self, message):
        if(not self.quiet):
            buffer = term_size().columns - len(message) - 3
            try:
                print('  ' + message + ' ' * buffer, end='\r', flush=True)
            except TypeError:
                print('  ' + message + ' ' * buffer, end='\r')

    def error(self, message):
        self.conwrite('')
        message = message.replace('\t', '    ')
        print('Error! {}'.format(message), file=sys.stderr)
        self.cleanup()

        from platform import system

        if(system() == 'Linux'):
            sys.exit(1)
        else:
            try:
                sys.exit(1)
            except SystemExit:
                import os
                os._exit(1)

    def bug(self, message, bug_type='bug report'):
        self.conwrite('')
        URL = 'https://github.com/WyattBlue/auto-editor/issues/'
        print('Error! {}\n\nSomething went wrong!\nCreate a {} at:\n  {}'.format(
            message, bug_type, URL), file=sys.stderr)
        self.cleanup()
        sys.exit(1)

    def warning(self, message):
        if(not self.quiet):
            print('Warning! {}'.format(message), file=sys.stderr)

    def print(self, message, end='\n'):
        if(not self.quiet):
            print(message, end=end)

    def checkType(self, data, name, correct_type):
        if(not isinstance(data, correct_type)):
            badtype = type(data).__name__
            goodtype = correct_type.__name__
            self.bug('Variable "{}" was not a {}, but a {}'.format(
                name, goodtype, badtype), 'bug report')
